# Wulfric - Cell, Atoms, K-path.
# Copyright (C) 2023-2025 Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from math import cos, sin, sqrt

import numpy as np

from wulfric.cell._basic_manipulation import from_params, get_reciprocal
from wulfric.constants import TORADIANS

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


# Primitive cell`s construction
def CUB(a: float):
    r"""
    Constructs primitive cubic cell as defined in [1]_.

    See :ref:`guide_cub` for the definition of primitive and conventional cells.

    Parameters
    ----------
    a : float or int
        Length of the three lattice vectors of the conventional cell.

    Returns
    -------
    cell : (3, 3) :numpy:`ndarray`
        Matrix of a primitive cell, rows are interpreted as vectors.

        .. code-block:: python

            cell = [[a1_x, a1_y, a1_z],
                    [a2_x, a2_y, a2_z],
                    [a3_x, a3_y, a3_z]]

    References
    ----------
    .. [1] Setyawan, W. and Curtarolo, S., 2010.
        High-throughput electronic band structure calculations: Challenges and tools.
        Computational materials science, 49(2), pp. 299-312.

    Examples
    --------

    .. doctest::

        >>> import wulfric as wulf
        >>> wulf.cell.CUB(a=2)
        array([[2, 0, 0],
               [0, 2, 0],
               [0, 0, 2]])
    """

    return np.array([[a, 0, 0], [0, a, 0], [0, 0, a]])


def FCC(a: float):
    r"""
    Constructs primitive face-centred cubic cell as defined in [1]_.

    See :ref:`guide_fcc` for the definition of primitive and conventional cells.

    Parameters
    ----------
    a : float
        Length of the three lattice vectors of the conventional cell.

    Returns
    -------
    cell : (3, 3) :numpy:`ndarray`
        Matrix of a primitive cell, rows are interpreted as vectors.

        .. code-block:: python

            cell = [[a1_x, a1_y, a1_z],
                    [a2_x, a2_y, a2_z],
                    [a3_x, a3_y, a3_z]]

    References
    ----------
    .. [1] Setyawan, W. and Curtarolo, S., 2010.
        High-throughput electronic band structure calculations: Challenges and tools.
        Computational materials science, 49(2), pp. 299-312.

    Examples
    --------

    .. doctest::

        >>> import wulfric as wulf
        >>> wulf.cell.FCC(a=2)
        array([[0., 1., 1.],
               [1., 0., 1.],
               [1., 1., 0.]])
    """

    return np.array([[0, a / 2, a / 2], [a / 2, 0, a / 2], [a / 2, a / 2, 0]])


def BCC(a: float):
    r"""
    Constructs primitive body-centred cubic cell as defined in [1]_.

    See :ref:`guide_bcc` for the definition of primitive and conventional cells.

    Parameters
    ----------
    a : float
        Length of the three lattice vectors of the conventional cell.

    Returns
    -------
    cell : (3, 3) :numpy:`ndarray`
        Matrix of a primitive cell, rows are interpreted as vectors.

        .. code-block:: python

            cell = [[a1_x, a1_y, a1_z],
                    [a2_x, a2_y, a2_z],
                    [a3_x, a3_y, a3_z]]

    References
    ----------
    .. [1] Setyawan, W. and Curtarolo, S., 2010.
        High-throughput electronic band structure calculations: Challenges and tools.
        Computational materials science, 49(2), pp. 299-312.

    Examples
    --------

    .. doctest::

        >>> import wulfric as wulf
        >>> wulf.cell.BCC(a=2)
        array([[-1.,  1.,  1.],
               [ 1., -1.,  1.],
               [ 1.,  1., -1.]])
    """

    return np.array(
        [[-a / 2, a / 2, a / 2], [a / 2, -a / 2, a / 2], [a / 2, a / 2, -a / 2]]
    )


def TET(a: float, c: float):
    r"""
    Constructs primitive tetragonal cell as defined in [1]_.

    See :ref:`guide_tet` for the definition of primitive and conventional cells.

    Parameters
    ----------
    a : float
        Length of the first two lattice vectors of the conventional cell.
    c : float
        Length of the third lattice vector of the conventional cell.

    Returns
    -------
    cell : (3, 3) :numpy:`ndarray`
        Matrix of a primitive cell, rows are interpreted as vectors.

        .. code-block:: python

            cell = [[a1_x, a1_y, a1_z],
                    [a2_x, a2_y, a2_z],
                    [a3_x, a3_y, a3_z]]

    References
    ----------
    .. [1] Setyawan, W. and Curtarolo, S., 2010.
        High-throughput electronic band structure calculations: Challenges and tools.
        Computational materials science, 49(2), pp. 299-312.

    Examples
    --------

    .. doctest::

        >>> import wulfric as wulf
        >>> wulf.cell.TET(a=2, c=5)
        array([[2, 0, 0],
               [0, 2, 0],
               [0, 0, 5]])
    """

    return np.array([[a, 0, 0], [0, a, 0], [0, 0, c]])


def BCT(a: float, c: float):
    r"""
    Constructs primitive body-centred tetragonal cell as defined in [1]_.

    See :ref:`guide_bct` for the definition of primitive and conventional cells.

    Parameters
    ----------
    a : float
        Length of the first two lattice vectors of the conventional cell.
    c : float
        Length of the third lattice vector of the conventional cell.

    Returns
    -------
    cell : (3, 3) :numpy:`ndarray`
        Matrix of a primitive cell, rows are interpreted as vectors.

        .. code-block:: python

            cell = [[a1_x, a1_y, a1_z],
                    [a2_x, a2_y, a2_z],
                    [a3_x, a3_y, a3_z]]

    References
    ----------
    .. [1] Setyawan, W. and Curtarolo, S., 2010.
        High-throughput electronic band structure calculations: Challenges and tools.
        Computational materials science, 49(2), pp. 299-312.

    Examples
    --------

    .. doctest::

        >>> import wulfric as wulf
        >>> wulf.cell.BCT(a=2, c=5)
        array([[-1. ,  1. ,  2.5],
               [ 1. , -1. ,  2.5],
               [ 1. ,  1. , -2.5]])
    """

    return np.array(
        [[-a / 2, a / 2, c / 2], [a / 2, -a / 2, c / 2], [a / 2, a / 2, -c / 2]]
    )


def ORC(a: float, b: float, c: float):
    r"""
    Constructs primitive orthorhombic cell as defined in [1]_.

    See :ref:`guide_orc` for the definition of primitive and conventional cells.

    Input values are used as they are, therefore, the cell might not be a standard
    primitive one.

    Parameters
    ----------
    a : float
        Length of the first lattice vector of the conventional cell.
    b : float
        Length of the second lattice vector of the conventional cell.
    c : float
        Length of the third lattice vector of the conventional cell.

    Returns
    -------
    cell : (3, 3) :numpy:`ndarray`
        Matrix of a primitive cell, rows are interpreted as vectors.

        .. code-block:: python

            cell = [[a1_x, a1_y, a1_z],
                    [a2_x, a2_y, a2_z],
                    [a3_x, a3_y, a3_z]]

    References
    ----------
    .. [1] Setyawan, W. and Curtarolo, S., 2010.
        High-throughput electronic band structure calculations: Challenges and tools.
        Computational materials science, 49(2), pp. 299-312.

    Examples
    --------

    .. doctest::

        >>> import wulfric as wulf
        >>> wulf.cell.ORC(a=3, b=5, c=7)
        array([[3, 0, 0],
               [0, 5, 0],
               [0, 0, 7]])
    """

    return np.array([[a, 0, 0], [0, b, 0], [0, 0, c]])


def ORCF(a: float, b: float, c: float):
    r"""
    Constructs primitive face-centred orthorhombic cell as defined in [1]_.

    See :ref:`guide_orcf` for the definition of primitive and conventional cells.

    Input values are used as they are, therefore, the cell might not be a standard
    primitive one.

    Parameters
    ----------
    a : float
        Length of the first lattice vector of the conventional cell.
    b : float
        Length of the second lattice vector of the conventional cell.
    c : float
        Length of the third lattice vector of the conventional cell.

    Returns
    -------
    cell : (3, 3) :numpy:`ndarray`
        Matrix of a primitive cell, rows are interpreted as vectors.

        .. code-block:: python

            cell = [[a1_x, a1_y, a1_z],
                    [a2_x, a2_y, a2_z],
                    [a3_x, a3_y, a3_z]]

    References
    ----------
    .. [1] Setyawan, W. and Curtarolo, S., 2010.
        High-throughput electronic band structure calculations: Challenges and tools.
        Computational materials science, 49(2), pp. 299-312.

    Examples
    --------

    .. doctest::

        >>> import wulfric as wulf
        >>> wulf.cell.ORCF(a=3, b=5, c=7)
        array([[0. , 2.5, 3.5],
               [1.5, 0. , 3.5],
               [1.5, 2.5, 0. ]])
    """

    return np.array([[0, b / 2, c / 2], [a / 2, 0, c / 2], [a / 2, b / 2, 0]])


def ORCI(a: float, b: float, c: float):
    r"""
    Constructs primitive body-centred orthorhombic cell as defined in [1]_.

    See :ref:`guide_orci` for the definition of primitive and conventional cells.

    Input values are used as they are, therefore, the cell might not be a standard
    primitive one.

    Parameters
    ----------
    a : float
        Length of the first lattice vector of the conventional cell.
    b : float
        Length of the second lattice vector of the conventional cell.
    c : float
        Length of the third lattice vector of the conventional cell.

    Returns
    -------
    cell : (3, 3) :numpy:`ndarray`
        Matrix of a primitive cell, rows are interpreted as vectors.

        .. code-block:: python

            cell = [[a1_x, a1_y, a1_z],
                    [a2_x, a2_y, a2_z],
                    [a3_x, a3_y, a3_z]]

    References
    ----------
    .. [1] Setyawan, W. and Curtarolo, S., 2010.
        High-throughput electronic band structure calculations: Challenges and tools.
        Computational materials science, 49(2), pp. 299-312.

    Examples
    --------

    .. doctest::

        >>> import wulfric as wulf
        >>> wulf.cell.ORCI(a=3, b=5, c=7)
        array([[-1.5,  2.5,  3.5],
               [ 1.5, -2.5,  3.5],
               [ 1.5,  2.5, -3.5]])
    """

    return np.array(
        [[-a / 2, b / 2, c / 2], [a / 2, -b / 2, c / 2], [a / 2, b / 2, -c / 2]]
    )


def ORCC(a: float, b: float, c: float):
    r"""
    Constructs primitive base-centred orthorhombic cell as defined in [1]_.

    See :ref:`guide_orcc` for the definition of primitive and conventional cells.

    Input values are used as they are, therefore, the cell might not be a standard
    primitive one.

    Parameters
    ----------
    a : float
        Length of the first lattice vector of the conventional cell.
    b : float
        Length of the second lattice vector of the conventional cell.
    c : float
        Length of the third lattice vector of the conventional cell.

    Returns
    -------
    cell : (3, 3) :numpy:`ndarray`
        Matrix of a primitive cell, rows are interpreted as vectors.

        .. code-block:: python

            cell = [[a1_x, a1_y, a1_z],
                    [a2_x, a2_y, a2_z],
                    [a3_x, a3_y, a3_z]]

    References
    ----------
    .. [1] Setyawan, W. and Curtarolo, S., 2010.
        High-throughput electronic band structure calculations: Challenges and tools.
        Computational materials science, 49(2), pp. 299-312.

    Examples
    --------

    .. doctest::

        >>> import wulfric as wulf
        >>> wulf.cell.ORCC(a=3, b=5, c=7)
        array([[ 1.5, -2.5,  0. ],
               [ 1.5,  2.5,  0. ],
               [ 0. ,  0. ,  7. ]])
    """

    return np.array([[a / 2, -b / 2, 0], [a / 2, b / 2, 0], [0, 0, c]])


def HEX(a: float, c: float):
    r"""
    Constructs primitive hexagonal cell as defined in [1]_.

    See :ref:`guide_hex` for the definition of primitive and conventional cells.

    Parameters
    ----------
    a : float
        Length of the first two lattice vectors of the conventional cell.
    c : float
        Length of the third lattice vector of the conventional cell.

    Returns
    -------
    cell : (3, 3) :numpy:`ndarray`
        Matrix of a primitive cell, rows are interpreted as vectors.

        .. code-block:: python

            cell = [[a1_x, a1_y, a1_z],
                    [a2_x, a2_y, a2_z],
                    [a3_x, a3_y, a3_z]]

    References
    ----------
    .. [1] Setyawan, W. and Curtarolo, S., 2010.
        High-throughput electronic band structure calculations: Challenges and tools.
        Computational materials science, 49(2), pp. 299-312.

    Examples
    --------

    .. doctest::

        >>> import wulfric as wulf
        >>> wulf.cell.HEX(a=3, c=5)
        array([[ 1.5       , -2.59807621,  0.        ],
               [ 1.5       ,  2.59807621,  0.        ],
               [ 0.        ,  0.        ,  5.        ]])
    """

    return np.array(
        [[a / 2, -a * sqrt(3) / 2, 0], [a / 2, a * sqrt(3) / 2, 0], [0, 0, c]]
    )


def RHL(a: float, alpha: float):
    r"""
    Constructs primitive rhombohedral cell as defined in [1]_.

    See :ref:`guide_rhl` for the definition of primitive and conventional cells.

    Input values are used as they are, therefore, the cell might not be a standard
    primitive one.

    Parameters
    ----------
    a : float
        Length of the lattice vectors of the conventional cell.
    alpha : float
        Angle between vectors :math:`a_2` and :math:`a_3` of the conventional cell in
        degrees.

    Returns
    -------
    cell : (3, 3) :numpy:`ndarray`
        Matrix of a primitive cell, rows are interpreted as vectors.

        .. code-block:: python

            cell = [[a1_x, a1_y, a1_z],
                    [a2_x, a2_y, a2_z],
                    [a3_x, a3_y, a3_z]]

    References
    ----------
    .. [1] Setyawan, W. and Curtarolo, S., 2010.
        High-throughput electronic band structure calculations: Challenges and tools.
        Computational materials science, 49(2), pp. 299-312.

    Examples
    --------

    .. doctest::

        >>> import wulfric as wulf
        >>> wulf.cell.RHL(a=3, alpha=40)
        array([[ 2.81907786, -1.02606043,  0.        ],
               [ 2.81907786,  1.02606043,  0.        ],
               [ 2.44562241,  0.        ,  1.73750713]])
    """

    alpha *= TORADIANS
    return np.array(
        [
            [a * cos(alpha / 2), -a * sin(alpha / 2), 0],
            [a * cos(alpha / 2), a * sin(alpha / 2), 0],
            [
                a * cos(alpha) / cos(alpha / 2),
                0,
                a * sqrt(1 - cos(alpha) ** 2 / cos(alpha / 2) ** 2),
            ],
        ]
    )


def MCL(a: float, b: float, c: float, alpha: float):
    r"""
    Constructs primitive monoclinic cell as defined in [1]_.

    See :ref:`guide_mcl` for the definition of primitive and conventional cells.

    Input values are used as they are, therefore, the cell might not be a standard
    primitive one.

    Parameters
    ----------
    a : float
        Length of the first lattice vector of the conventional cell.
    b : float
        Length of the second of the two remaining lattice vectors of the conventional
        cell.
    c : float
        Length of the third of the two remaining lattice vectors of the conventional cell.
    alpha : float
        Angle between vectors :math:`a_2` and :math:`a_3` of the conventional cell in
        degrees.

    Returns
    -------
    cell : (3, 3) :numpy:`ndarray`
        Matrix of a primitive cell, rows are interpreted as vectors.

        .. code-block:: python

            cell = [[a1_x, a1_y, a1_z],
                    [a2_x, a2_y, a2_z],
                    [a3_x, a3_y, a3_z]]

    References
    ----------
    .. [1] Setyawan, W. and Curtarolo, S., 2010.
        High-throughput electronic band structure calculations: Challenges and tools.
        Computational materials science, 49(2), pp. 299-312.

    Examples
    --------

    .. doctest::

        >>> import wulfric as wulf
        >>> wulf.cell.MCL(a=3, b=5, c=7, alpha = 45)
        array([[3.        , 0.        , 0.        ],
               [0.        , 5.        , 0.        ],
               [0.        , 4.94974747, 4.94974747]])
    """

    alpha *= TORADIANS
    return np.array([[a, 0, 0], [0, b, 0], [0, c * cos(alpha), c * sin(alpha)]])


def MCLC(a: float, b: float, c: float, alpha: float):
    r"""
    Constructs primitive base-centred monoclinic cell as defined in [1]_.

    See :ref:`guide_mclc` for the definition of primitive and conventional cells.

    Input values are used as they are, therefore, the cell might not be a standard
    primitive one.

    Parameters
    ----------
    a : float
        Length of the first lattice vector of the conventional cell.
    b : float
        Length of the second of the two remaining lattice vectors of the conventional
        cell.
    c : float
        Length of the third of the two remaining lattice vectors of the conventional
        cell.
    alpha : float
        Angle between vectors :math:`a_2` and :math:`a_3` of the conventional cell in
        degrees.

    Returns
    -------
    cell : (3, 3) :numpy:`ndarray`
        Matrix of a primitive cell, rows are interpreted as vectors.

        .. code-block:: python

            cell = [[a1_x, a1_y, a1_z],
                    [a2_x, a2_y, a2_z],
                    [a3_x, a3_y, a3_z]]

    References
    ----------
    .. [1] Setyawan, W. and Curtarolo, S., 2010.
        High-throughput electronic band structure calculations: Challenges and tools.
        Computational materials science, 49(2), pp. 299-312.

    Examples
    --------

    .. doctest::

        >>> import wulfric as wulf
        >>> wulf.cell.MCLC(a=3, b=5, c=7, alpha = 45)
        array([[ 1.5       ,  2.5       ,  0.        ],
               [-1.5       ,  2.5       ,  0.        ],
               [ 0.        ,  4.94974747,  4.94974747]])
    """

    alpha *= TORADIANS
    return np.array(
        [
            [a / 2, b / 2, 0],
            [-a / 2, b / 2, 0],
            [0, c * cos(alpha), c * sin(alpha)],
        ]
    )


def TRI(
    a: float,
    b: float,
    c: float,
    alpha: float,
    beta: float,
    gamma: float,
    input_reciprocal=False,
):
    r"""
    Constructs primitive triclinic cell as defined in [1]_.

    See :ref:`guide_tri` for the definition of primitive and conventional cells.

    Parameters
    ----------
    a : float
        Length of the first lattice vector of the conventional cell.
    b : float
        Length of the second lattice vector of the conventional cell.
    c : float
        Length of the third lattice vector of the conventional cell.
    alpha : float
        Angle between vectors :math:`a_2` and :math:`a_3` of the conventional cell in
        degrees.
    beta : float
        Angle between vectors :math:`a_1` and :math:`a_3` of the conventional cell in
        degrees.
    gamma : float
        Angle between vectors :math:`a_1` and :math:`a_2` of the conventional cell in
        degrees.
    input_reciprocal : bool, default False
        Whether to interpret input as reciprocal parameters.

    Returns
    -------
    cell : (3, 3) :numpy:`ndarray`
        Matrix of a primitive cell, rows are interpreted as vectors.

        .. code-block:: python

            cell = [[a1_x, a1_y, a1_z],
                    [a2_x, a2_y, a2_z],
                    [a3_x, a3_y, a3_z]]

    References
    ----------
    .. [1] Setyawan, W. and Curtarolo, S., 2010.
        High-throughput electronic band structure calculations: Challenges and tools.
        Computational materials science, 49(2), pp. 299-312.

    Examples
    --------

    .. doctest::

        >>> import wulfric as wulf
        >>> wulf.cell.TRI(a=3, b=5, c=7, alpha = 45, beta = 33, gamma = 21)
        array([[ 3.        ,  0.        ,  0.        ],
               [ 4.66790213,  1.79183975,  0.        ],
               [ 5.87069398, -1.48176621,  3.51273699]])
    """

    cell = from_params(a, b, c, alpha, beta, gamma)
    if input_reciprocal:
        cell = get_reciprocal(cell)

    return cell


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# Wulfric - Cell, Atoms, K-path.
# Copyright (C) 2023-2025 Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import numpy as np

from wulfric.geometry._geometry import absolute_to_relative

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


def shift_atoms(
    atoms, gravity_point=(0.5, 0.5, 0.5), cell=None, gp_is_relative=True
) -> None:
    R"""
    Shifts all atoms with the same vector in a way
    that the ``gravity_point`` is located in the middle between minimum and maximum
    relative coordinates of the atoms, individually for each lattice vector.

    I.e. if there is one atom in the cell, then it is placed in the center of the cell
    for ``gravity_point`` = (0.5, 0.5, 0.5).

    Modifies given ``atoms`` dictionary.

    Parameters
    ----------
    atoms : dict
        Dictionary with atoms. Must have a ``"positions"`` with value of (N,3) |array-like|_.
    gravity_point : (3,) |array-like|_, default (0.5, 0.5, 0.5)
        Relative coordinates of the gravity point.
    cell : (3, 3) |array-like|_, optional
        Matrix of a cell, rows are interpreted as vectors. Required if
        ``gp_is_relative = False``.
    gp_is_relative : bool, default True
        Whether the ``gravity_point`` is given in relative coordinates.

    Examples
    --------

    .. doctest::

        >>> import wulfric as wulf
        >>> cell=[[2, 0, 0], [0, 2, 0], [0, 0, 2]]
        >>> atoms = {"names" : ['Cr1', 'Cr2'], "positions" : [[0.0, 0.0, 0.0], [0.5, 0.5, 1.0]]}
        >>> wulf.crystal.shift_atoms(atoms=atoms, gravity_point=(0.5, 0.5, 0.5))
        >>> for i in range(len(atoms["names"])):
        ...    print(atoms["names"][i], atoms["positions"][i])
        ...
        Cr1 [0.25 0.25 0.  ]
        Cr2 [0.75 0.75 1.  ]
        >>> wulf.crystal.shift_atoms(atoms, gravity_point=(1,1,1), cell=cell, gp_is_relative=False)
        >>> for i in range(len(atoms["names"])):
        ...    print(atoms["names"][i], atoms["positions"][i])
        ...
        Cr1 [0.25 0.25 0.  ]
        Cr2 [0.75 0.75 1.  ]
    """

    if not gp_is_relative:
        if cell is None:
            raise ValueError("cell is required if gp_is_relative False")

        gravity_point = absolute_to_relative(gravity_point, cell)

    min_coord = np.min(atoms["positions"], axis=0)
    max_coord = np.max(atoms["positions"], axis=0)
    shift = (max_coord + min_coord) / 2

    atoms["positions"] = [
        position - shift + gravity_point for position in atoms["positions"]
    ]


def cure_negative(atoms) -> None:
    R"""
    Shifts all atoms with the same vector in a way
    that all relative coordinates becomes non-negative.

    Modifies given ``atoms`` dictionary.

    Parameters
    ----------
    atoms : dict
        Dictionary with atoms. Must have a ``"positions"`` with value of (N,3) |array-like|_.

    Examples
    --------

    .. doctest::

        >>> import wulfric as wulf
        >>> cell=[[2, 0, 0], [0, 2, 0], [0, 0, 2]]
        >>> atoms = {"names" : ['Cr1', 'Cr2'], "positions" : [[-0.5, 0.5, 0.0], [0.1, 0.5, 0.0]] }
        >>> wulf.crystal.cure_negative(atoms)
        >>> for i in range(len(atoms["names"])):
        ...    print(atoms["names"][i], atoms["positions"][i])
        ...
        Cr1 [0.  0.5 0. ]
        Cr2 [0.6 0.5 0. ]
    """

    min_values = atoms["positions"][0]
    for position in atoms["positions"][1:]:
        min_values = np.minimum(min_values, position)

    shift = np.where(min_values < 0, -min_values, 0)

    atoms["positions"] = [position + shift for position in atoms["positions"]]


def ensure_000(atoms) -> None:
    r"""
    Ensures that all atoms are from (0,0,0) unit cell.

    In other word ensures that all relative coordinates of all atoms are :math:`\in [0,1]`.input

    Parameters
    ----------
    atoms : dict
        Dictionary with atoms. Must have a ``"positions"`` with value of (N,3) |array-like|_.

    Examples
    --------

    .. doctest::

        >>> import wulfric as wulf
        >>> atoms = {"positions" : [[0, 0.5, 0], [1.25, 0, -0.52], [0.25, -0.65, 2.375]]}
        >>> for p in atoms["positions"]:
        ...        print(p)
        ...
        [0, 0.5, 0]
        [1.25, 0, -0.52]
        [0.25, -0.65, 2.375]
        >>> wulf.crystal.ensure_000(atoms)
        >>> for p in atoms["positions"]:
        ...        print(p)
        ...
        [0, 0.5, 0]
        [0.25, 0, 0.48]
        [0.25, 0.35, 0.375]
    """

    for i in range(len(atoms["positions"])):
        for j in range(3):
            poscomp = atoms["positions"][i][j]

            # Ensure -1 < poscomp < 1
            poscomp -= int(poscomp)

            # Ensure 0 <= poscomp <= 1
            if poscomp < 0:
                poscomp += 1

            atoms["positions"][i][j] = poscomp


def get_vector(
    cell, atoms, atom1, atom2, R=(0, 0, 0), return_relative=False
) -> np.ndarray:
    r"""
    Computes a vector from atom1 (from (0,0,0)) to atom2 (from (i,j,k)).

    Parameters
    ----------
    cell : (3, 3) |array-like|_,
        Matrix of a cell, rows are interpreted as vectors.
    atoms : dict
        Dictionary with atoms. Must have a ``"positions"`` with value of (N,3) |array-like|_.
    atom1 : int
        Index of the first atom in ``atoms["positions"]``.
    atom2 : int
        Index of the second atom in ``atoms["positions"]``.
    R : (3,) tuple of int, default (0, 0, 0)
        Radius vector of the unit cell for atom2 (i,j,k).
    return_relative : bool, default False
        Whether to return vector relative to the ``cell``.

    Returns
    -------
    v : (3,) :numpy:`ndarray`
        Vector from atom1 in (0,0,0) cell to atom2 in R cell.

    Examples
    --------

    .. doctest::

        >>> import wulfric as wulf
        >>> cell = [[1, 0, 0], [0, 2, 0], [0, 0, 3]]
        >>> atoms = {"positions" : [[0.5, 0, 0], [0, 0, 0.5]]}
        >>> wulf.crystal.get_vector(cell, atoms, atom1=0, atom2=1, R=(0, 0, 0))
        array([-0.5,  0. ,  1.5])
        >>> wulf.crystal.get_vector(cell, atoms, atom1=0, atom2=1, R=(1, 0, 0))
        array([0.5, 0. , 1.5])
        >>> wulf.crystal.get_vector(cell, atoms, atom1=0, atom2=1, R=(1, 0, -3))
        array([ 0.5,  0. , -7.5])
        >>> wulf.crystal.get_vector(cell, atoms, atom1=0, atom2=1, R=(1, 0, -3), return_relative=True)
        array([ 0.5,  0. , -2.5])
    """

    relative_vector = (
        np.array(R, dtype=float) + atoms["positions"][atom2] - atoms["positions"][atom1]
    )

    if return_relative:
        return relative_vector

    return relative_vector @ cell


def get_distance(cell, atoms, atom1, atom2, R=(0, 0, 0)) -> float:
    r"""
    Computes distance between atom1 (from (0,0,0)) and atom2 (from (i,j,k)).

    Parameters
    ----------
    cell : (3, 3) |array-like|_,
        Matrix of a cell, rows are interpreted as vectors.
    atoms : dict
        Dictionary with atoms. Must have a ``"positions"`` with value of (N,3) |array-like|_.
    atom1 : int
        Index of the first atom in ``atoms["positions"]``.
    atom2 : int
        Index of the second atom in ``atoms["positions"]``.
    R : (3,) tuple of int, default (0, 0, 0)
        Radius vector of the unit cell for atom2 (i,j,k).

    Returns
    -------
    distance : float
        Distance between atom1 in (0,0,0) cell and atom2 in R cell.

    Examples
    --------

    .. doctest::

        >>> import wulfric as wulf
        >>> cell = [[1, 0, 0], [0, 2, 0], [0, 0, 3]]
        >>> atoms = {"positions" : [[0.5, 0, 0], [0, 0, 0.5]]}
        >>> round(wulf.crystal.get_distance(cell, atoms, atom1=0, atom2=1, R=(0, 0, 0)), 8)
        1.58113883
        >>> round(wulf.crystal.get_distance(cell, atoms, atom1=0, atom2=1, R=(1, 0, 0)), 8)
        1.58113883
        >>> round(wulf.crystal.get_distance(cell, atoms, atom1=0, atom2=1, R=(1, 0, -3)), 8)
        7.51664819
    """

    return float(
        np.linalg.norm(
            get_vector(
                cell=cell,
                atoms=atoms,
                atom1=atom1,
                atom2=atom2,
                R=R,
                return_relative=False,
            )
        )
    )


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# Wulfric - Cell, Atoms, K-path.
# Copyright (C) 2023-2025 Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from math import cos, sin

import numpy as np

from wulfric._numerical import compare_numerically
from wulfric.cell._basic_manipulation import get_params, get_reciprocal
from wulfric.cell._lepage import lepage
from wulfric.cell._sc_standardize import get_conventional
from wulfric.constants._numerical import TORADIANS
from wulfric.geometry._geometry import get_volume

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


def _BCT_variation(conv_a: float, conv_c: float):
    r"""
    Two variations of the BCT lattice.

    Condition :math:`a \ne c` is assumed.

    :math:`\text{BCT}_1: c < a` and :math:`\text{BCT}_2: c > a`

    Parameters
    ----------
    conv_a : float
        Length of the :math:`a_1 == a_2` vector of the conventional cell.
    conv_c : float
        Length of the :math:`a_3` vector of the conventional cell.

    Returns
    -------
    variation : str
        Variation of the lattice. "BCT1" or "BCT2".

    Raises
    ------
    ValueError
        If :math:`a == c`.
    """
    if conv_a > conv_c:
        return "BCT1"
    elif conv_a < conv_c:
        return "BCT2"
    else:
        raise ValueError("a == c")


def _ORCF_variation(conv_a: float, conv_b: float, conv_c: float, length_tolerance=1e-8):
    r"""
    Three variations of the ORCF lattice.

    Ordering :math:`a < b < c` is assumed.

    :math:`\text{ORCF}_1: \dfrac{1}{a^2} > \dfrac{1}{b^2} + \dfrac{1}{c^2}`,
    :math:`\text{ORCF}_2: \dfrac{1}{a^2} < \dfrac{1}{b^2} + \dfrac{1}{c^2}`,
    :math:`\text{ORCF}_3: \dfrac{1}{a^2} = \dfrac{1}{b^2} + \dfrac{1}{c^2}`,

    Parameters
    ----------
    conv_a : float
        Length of the :math:`a_1` vector of the conventional cell.
    conv_b : float
        Length of the :math:`a_2` vector of the conventional cell.
    conv_c : float
        Length of the :math:`a_3` vector of the conventional cell.
    length_tolerance : float, default :math:`10^{-8}`
        Tolerance for length variables (lengths of the lattice vectors). Default value is
        chosen in the contexts of condense matter physics, assuming that length is given
        in Angstroms. Please choose appropriate tolerance for your problem.

    Returns
    -------
    variation : str
        Variation of the lattice. "ORCF1", "ORCF2" or "ORCF3".

    Raises
    ------
    ValueError
        If :math:`a < b < c` is not satisfied.
    """
    if compare_numerically(
        conv_a, ">=", conv_b, eps=length_tolerance
    ) or compare_numerically(conv_b, ">=", conv_c, eps=length_tolerance):
        raise ValueError(
            f"a < b < c is not satisfied with {length_tolerance} tolerance."
        )

    expression = 1 / conv_a**2 - 1 / conv_b**2 - 1 / conv_c**2
    if compare_numerically(expression, "==", 0, eps=length_tolerance):
        return "ORCF3"
    elif compare_numerically(expression, ">", 0, eps=length_tolerance):
        return "ORCF1"
    elif compare_numerically(expression, "<", 0, eps=length_tolerance):
        return "ORCF2"


def _RHL_variation(conv_alpha: float, angle_tolerance=1e-4):
    r"""
    Two variations of the RHL lattice.

    Condition :math:`\alpha \ne 90^{\circ}` is assumed.

    :math:`\text{RHL}_1 \alpha < 90^{\circ}`,
    :math:`\text{RHL}_2 \alpha > 90^{\circ}`

    Parameters
    ----------
    conv_alpha : float
        Angle between vectors :math:`a_1` and :math:`a_2` of the conventional cell in
        degrees.
    angle_tolerance : float, default :math:`10^{-4}`
        Tolerance for angle variables (angles of the lattice). Default value is chosen in
        the contexts of condense matter physics, assuming that angles are in degrees.
        Please choose appropriate tolerance for your problem.

    Returns
    -------
    variation : str
        Variation of the lattice. Either "RHL1" or "RHL2".

    Raises
    ------
    ValueError
        If :math:`\alpha == 90^{\circ}` with given tolerance ``eps``.
    """
    if compare_numerically(conv_alpha, "<", 90, eps=angle_tolerance):
        return "RHL1"
    elif compare_numerically(conv_alpha, ">", 90, eps=angle_tolerance):
        return "RHL2"
    else:
        raise ValueError(f"alpha == 90 with {angle_tolerance} tolerance.")


def _MCLC_variation(
    conv_a: float,
    conv_b: float,
    conv_c: float,
    conv_alpha: float,
    k_gamma: float,
    length_tolerance=1e-8,
    angle_tolerance=1e-4,
):
    r"""
    Five variation of the MCLC lattice.

    Ordering :math:`a \le c` and :math:`b \le c` and :math:`\alpha < 90^{\circ}` is assumed.

    :math:`\text{MCLC}_1: k_{\gamma} > 90^{\circ}`,
    :math:`\text{MCLC}_2: k_{\gamma} = 90^{\circ}`,
    :math:`\text{MCLC}_3: k_{\gamma} < 90^{\circ}, \dfrac{b\cos(\alpha)}{c} + \dfrac{b^2\sin(\alpha)^2}{a^2} < 1`
    :math:`\text{MCLC}_4: k_{\gamma} < 90^{\circ}, \dfrac{b\cos(\alpha)}{c} + \dfrac{b^2\sin(\alpha)^2}{a^2} = 1`
    :math:`\text{MCLC}_5: k_{\gamma} < 90^{\circ}, \dfrac{b\cos(\alpha)}{c} + \dfrac{b^2\sin(\alpha)^2}{a^2} > 1`

    Parameters
    ----------
    conv_a : float
        Length of the :math:`a_1` vector of the conventional cell.
    conv_b : float
        Length of the :math:`a_2` vector of the conventional cell.
    conv_c : float
        Length of the :math:`a_3` vector of the conventional cell.
    conv_alpha : float
        Angle between vectors :math:`a_2` and :math:`a_3` of the conventional cell in
        degrees.
    k_gamma : float
        Angle between reciprocal vectors :math:`b_1` and :math:`b_2`. In degrees.
    length_tolerance : float, default :math:`10^{-8}`
        Tolerance for length variables (lengths of the lattice vectors). Default value is
        chosen in the contexts of condense matter physics, assuming that length is given
        in Angstroms. Please choose appropriate tolerance for your problem.
    angle_tolerance : float, default :math:`10^{-4}`
        Tolerance for angle variables (angles of the lattice). Default value is chosen in
        the contexts of condense matter physics, assuming that angles are in degrees.
        Please choose appropriate tolerance for your problem.

    Returns
    -------
    variation : str
        Variation of the lattice.
        Either "MCLC1", "MCLC2", "MCLC3", "MCLC4" or "MCLC5".

    Raises
    ------
    ValueError
        If :math:`\alpha > 90^{\circ}` or :math:`a > c` or :math:`b > c` with given
        tolerance ``eps``.
    """

    if compare_numerically(
        conv_alpha, ">", 90, eps=angle_tolerance
    ) or compare_numerically(conv_b, ">", conv_c, eps=length_tolerance):
        raise ValueError(
            f"alpha > 90 or or b > c with {angle_tolerance} or {length_tolerance} tolerance:\n"
            + f"  alpha = {conv_alpha}\n"
            + f"  b = {conv_b}\n"
            + f"  c = {conv_c}"
        )

    conv_alpha *= TORADIANS

    if compare_numerically(k_gamma, "==", 90, eps=angle_tolerance):
        return "MCLC2"
    elif compare_numerically(k_gamma, ">", 90, eps=angle_tolerance):
        return "MCLC1"
    elif compare_numerically(k_gamma, "<", 90, eps=angle_tolerance):
        expression = (
            conv_b * cos(conv_alpha) / conv_c
            + conv_b**2 * sin(conv_alpha) ** 2 / conv_a**2
        )
        if compare_numerically(expression, "==", 1, eps=length_tolerance):
            return "MCLC4"
        elif compare_numerically(expression, "<", 1, eps=length_tolerance):
            return "MCLC3"
        elif compare_numerically(expression, ">", 1, eps=length_tolerance):
            return "MCLC5"


def _TRI_variation(k_alpha: float, k_beta: float, k_gamma: float, angle_tolerance=1e-4):
    r"""
    Four variations of the TRI lattice.

    Conditions :math:`k_{\alpha} \ne 90^{\circ}` and :math:`k_{\beta} \ne 90^{\circ}` are assumed.

    :math:`\text{TRI}_{1a} k_{\alpha} > 90^{\circ}, k_{\beta} > 90^{\circ}, k_{\gamma} > 90^{\circ}, k_{\gamma} = \min(k_{\alpha}, k_{\beta}, k_{\gamma})`

    :math:`\text{TRI}_{1b} k_{\alpha} < 90^{\circ}, k_{\beta} < 90^{\circ}, k_{\gamma} < 90^{\circ}, k_{\gamma} = \max(k_{\alpha}, k_{\beta}, k_{\gamma})`

    :math:`\text{TRI}_{2a} k_{\alpha} > 90^{\circ}, k_{\beta} > 90^{\circ}, k_{\gamma} = 90^{\circ}`

    :math:`\text{TRI}_{2b} k_{\alpha} < 90^{\circ}, k_{\beta} < 90^{\circ}, k_{\gamma} = 90^{\circ}`

    Parameters
    ----------
    k_alpha : float
        Angle between reciprocal vectors :math:`b_2` and :math:`b_3`. In degrees.
    k_beta : float
        Angle between reciprocal vectors :math:`b_1` and :math:`b_3`. In degrees.
    k_gamma : float
        Angle between reciprocal vectors :math:`b_1` and :math:`b_2`. In degrees.
    angle_tolerance : float, default :math:`10^{-4}`
        Tolerance for angle variables (angles of the lattice). Default value is chosen in
        the contexts of condense matter physics, assuming that angles are in degrees.
        Please choose appropriate tolerance for your problem.

    Returns
    -------
    variation : str
        Variation of the lattice.
        Either "TRI1a", "TRI1b", "TRI2a" or "TRI2b".

    Raises
    ------
    ValueError
        If :math:`k_{\alpha} == 90^{\circ}` or :math:`k_{\beta} == 90^{\circ}` with given
        tolerance ``eps``.
    """

    if compare_numerically(
        k_alpha, "==", 90, eps=angle_tolerance
    ) or compare_numerically(k_beta, "==", 90, eps=angle_tolerance):
        raise ValueError(
            f"k_alpha == 90 or k_beta == 90 with {angle_tolerance} tolerance."
        )

    if compare_numerically(k_gamma, "==", 90, eps=angle_tolerance):
        if compare_numerically(
            k_alpha, ">", 90, eps=angle_tolerance
        ) and compare_numerically(k_beta, ">", 90, eps=angle_tolerance):
            return "TRI2a"
        elif compare_numerically(
            k_alpha, "<", 90, eps=angle_tolerance
        ) and compare_numerically(k_beta, "<", 90, eps=angle_tolerance):
            return "TRI2b"
    elif compare_numerically(
        min(k_gamma, k_beta, k_alpha), ">", 90, eps=angle_tolerance
    ):
        return "TRI1a"
    elif compare_numerically(
        max(k_gamma, k_beta, k_alpha), "<", 90, eps=angle_tolerance
    ):
        return "TRI1b"
    else:
        return "TRI"


def get_variation(cell, lattice_type=None, length_tolerance=1e-8, angle_tolerance=1e-4):
    r"""
    Return variation of the lattice as defined in the paper by Setyawan and Curtarolo [1]_.

    Parameters
    ----------
    cell : (3, 3) |array-like|_
        Matrix of a primitive cell, rows are interpreted as vectors.
    lattice_type : str, optional
        One of the 14 lattice types that correspond to the provided ``cell``,
        case-insensitive. If not provided, then computed automatically from ``cell``. If
        provided, then it user's responsibility to ensure that ``lattice_type`` is
        correct.
    length_tolerance : float, default :math:`10^{-8}`
        Tolerance for length variables (lengths of the lattice vectors). Default value is
        chosen in the contexts of condense matter physics, assuming that length is given
        in Angstroms. Please choose appropriate tolerance for your problem.
    angle_tolerance : float, default :math:`10^{-4}`
        Tolerance for angle variables (angles of the lattice). Default value is chosen in
        the contexts of condense matter physics, assuming that angles are in degrees.
        Please choose appropriate tolerance for your problem.

    Returns
    -------
    variation : str
        Variation of the lattice defined by the ``cell``.

    References
    ----------
    .. [1] Setyawan, W. and Curtarolo, S., 2010.
        High-throughput electronic band structure calculations: Challenges and tools.
        Computational materials science, 49(2), pp. 299-312.

    Examples
    --------

    .. doctest::

        >>> import wulfric as wulf
        >>> # There is no variation of cubic lattice, therefore, just lattice type is
        >>> # returned
        >>> wulf.cell.get_variation([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
        'CUB'
        >>> cell = wulf.cell.get_cell_example("bct")
        >>> wulf.cell.get_variation(cell)
        'BCT1'
        >>> # If lattice type is given and it is incorrect, then the behaviour is undefined.abs
        >>> # It may return wrong result
        >>> wulf.cell.get_variation(cell, lattice_type="cub")
        'CUB'
        >>> # Or give an error
        >>> wulf.cell.get_variation(cell, lattice_type="mclc")
        Traceback (most recent call last):
        ...
        ValueError: alpha > 90 or or b > c with 0.0001 or 1e-08 tolerance:
        ...

    """

    cell = np.array(cell, dtype=float)

    if lattice_type is None:
        lattice_type = lepage(cell, angle_tolerance=angle_tolerance)

    lattice_type = lattice_type.upper()

    if lattice_type in ["BCT", "ORCF", "RHL", "MCLC"]:
        conv_a, conv_b, conv_c, conv_alpha, conv_beta, conv_gamma = get_params(
            get_conventional(cell)
        )

    if lattice_type == "BCT":
        result = _BCT_variation(conv_a, conv_c)
    elif lattice_type == "ORCF":
        result = _ORCF_variation(
            conv_a, conv_b, conv_c, length_tolerance=length_tolerance
        )
    elif lattice_type == "RHL":
        result = _RHL_variation(conv_alpha, angle_tolerance=angle_tolerance)
    elif lattice_type == "MCLC":
        _, _, _, _, _, k_gamma = get_params(get_reciprocal(cell))
        result = _MCLC_variation(
            conv_a,
            conv_b,
            conv_c,
            conv_alpha,
            k_gamma,
            length_tolerance=length_tolerance,
            angle_tolerance=angle_tolerance,
        )
    elif lattice_type == "TRI":
        _, _, _, k_alpha, k_beta, k_gamma = get_params(get_reciprocal(cell))
        result = _TRI_variation(
            k_alpha,
            k_beta,
            k_gamma,
            angle_tolerance=angle_tolerance,
        )
    else:
        result = lattice_type

    return result


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# Wulfric - Cell, Atoms, K-path.
# Copyright (C) 2023-2025 Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import numpy as np

try:
    from scipy.spatial import Voronoi

    SCIPY_AVAILABLE = True
except ImportError:
    SCIPY_AVAILABLE = False

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


def _lattice_points(cell, relative=False):
    r"""
    Compute lattice points

    Parameters
    ----------
    cell : (3, 3) |array-like|_
        Matrix of a cell, rows are interpreted as vectors.
    relative : bool, default False
        Whether to return relative coordinates.

    Returns
    -------
    lattice_points : (N, 3) :numpy:`ndarray`
        N lattice points. Each element is a vector :math:`v = (v_x, v_y, v_z)`.
    """

    lattice_points = np.zeros((27, 3), dtype=float)
    for i in [-1, 0, 1]:
        for j in [-1, 0, 1]:
            for k in [-1, 0, 1]:
                point = np.array([i, j, k])
                if not relative:
                    point = point @ cell
                lattice_points[9 * (i + 1) + 3 * (j + 1) + (k + 1)] = point
    return lattice_points


def _get_voronoi_cell(cell):
    r"""
    Computes Voronoi edges around (0,0,0) point.

    Parameters
    ----------
    cell : (3, 3) |array-like|_
        Matrix of a cell, rows are interpreted as vectors.

    Returns
    -------
    edges : (N, 2, 3) :numpy:`ndarray`
        N edges of the Voronoi cell around (0,0,0) point. Each elements contains two
        vectors of the points of the voronoi vertices forming an edge.
    vertices : (M, 3) :numpy:`ndarray`
        M vertices of the Voronoi cell around (0,0,0) point. Each element is a vector
        :math:`v = (v_x, v_y, v_z)`.

    Notes
    -----
    This function is a part of ``wulfric[visual]``
    """

    if not SCIPY_AVAILABLE:
        raise ImportError('SciPy is not available. Install it with "pip install scipy"')
    voronoi = Voronoi(_lattice_points(cell, relative=False))
    edges_index = set()
    # Thanks ase for the idea. 13 - is the index of (0,0,0) point.
    for rv, rp in zip(voronoi.ridge_vertices, voronoi.ridge_points):
        if -1 not in rv and 13 in rp:
            for j in range(0, len(rv)):
                if (rv[j - 1], rv[j]) not in edges_index and (
                    rv[j],
                    rv[j - 1],
                ) not in edges_index:
                    edges_index.add((rv[j - 1], rv[j]))
    edges_index = np.array(list(edges_index))
    edges = np.zeros((edges_index.shape[0], 2, 3), dtype=voronoi.vertices.dtype)
    for i in range(edges_index.shape[0]):
        edges[i][0] = voronoi.vertices[edges_index[i][0]]
        edges[i][1] = voronoi.vertices[edges_index[i][1]]
    return edges, voronoi.vertices[np.unique(edges_index.flatten())]


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# Wulfric - Cell, Atoms, K-path.
# Copyright (C) 2023-2025 Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from math import cos
from math import pi as PI
from math import sin, sqrt

import numpy as np

from wulfric.constants._numerical import TORADIANS
from wulfric.geometry._geometry import get_angle, get_volume, parallelepiped_check

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


def get_reciprocal(cell):
    r"""
    Computes reciprocal cell.

    Parameters
    ----------
    cell : (3, 3) |array-like|_
        Matrix of a direct cell, rows are interpreted as vectors.

    Returns
    -------
    reciprocal_cell : (3, 3) :numpy:`ndarray`
        Matrix of a reciprocal cell.

        .. code-block:: python

            cell = [[b1_x, b1_y, b1_z],
                    [b2_x, b2_y, b2_z],
                    [b3_x, b3_y, b3_z]]

    Examples
    --------

    .. doctest::

        >>> import wulfric as wulf
        >>> cell = [[1, 1, 0], [0, 1, 0], [0, 0, 1]]
        >>> wulf.cell.get_reciprocal(cell)
        array([[ 6.28318531,  0.        ,  0.        ],
               [-6.28318531,  6.28318531,  0.        ],
               [ 0.        ,  0.        ,  6.28318531]])

    """

    return 2 * PI * np.linalg.inv(cell).T


def from_params(a=1.0, b=1.0, c=1.0, alpha=90.0, beta=90.0, gamma=90.0):
    r"""
    Constructs the cell from lattice parameters.

    The spatial orientation of the cell is decided as

    *   Lattice vector :math:`\boldsymbol{a_1}` has the length ``a`` and oriented along
        :math:`{\cal x}` axis.
    *   Lattice vector :math:`\boldsymbol{a_2}` has the length ``b``, is placed in
        :math:`{\cal xy}` plane and form an angle ``gamma`` with vector
        :math:`\boldsymbol{a_1}`, positive in a mathematical sense.
    *   Lattice vector :math:`\boldsymbol{a_3}` has the length ``c`` and form an angle
        ``alpha`` with the vector :math:`\boldsymbol{a_2}` and an angle ``beta`` with
        the vector :math:`\boldsymbol{a_1}`.

    Parameters
    ----------
    a : float, default 1.0
        Length of the :math:`\boldsymbol{a_1}` vector.
    b : float, default 1.0
        Length of the :math:`\boldsymbol{a_2}` vector.
    c : float, default 1.0
        Length of the :math:`\boldsymbol{a_3}` vector.
    alpha : float, default 90.0
        Angle between vectors :math:`\boldsymbol{a_2}` and :math:`\boldsymbol{a_3}` given
        in degrees.
    beta : float, default 90.0
        Angle between vectors :math:`\boldsymbol{a_1}` and :math:`\boldsymbol{a_3}` given
        in degrees.
    gamma : float, default 90.0
        Angle between vectors :math:`\boldsymbol{a_1}` and :math:`\boldsymbol{a_2}` given
        in degrees.

    Returns
    -------
    cell : (3, 3) :numpy:`ndarray`
        Matrix of a direct cell, rows are interpreted as vectors.

        .. code-block:: python

            cell = [[a1_x, a1_y, a1_z],
                    [a2_x, a2_y, a2_z],
                    [a3_x, a3_y, a3_z]]

    Raises
    ------
    ValueError
        If parameters could not form a parallelepiped.

    See Also
    --------
    get_params
    wulfric.geometry.parallelepiped_check : Check if parameters can form a parallelepiped.

    Examples
    --------

    .. doctest::

        >>> import wulfric as wulf
        >>> import numpy as np
        >>> np.around(wulf.cell.from_params(1, 2, 3, 90, 90, 60), 6)
        array([[1.      , 0.      , 0.      ],
               [1.      , 1.732051, 0.      ],
               [0.      , 0.      , 3.      ]])

    .. doctest::

        >>> wulf.cell.from_params(1, 2, 3, 60, 60, 130)
        Traceback (most recent call last):
        ...
        ValueError: Parameters could not form a parallelepiped:
        a = 1
        b = 2
        c = 3
        alpha = 60
        beta = 60
        gamma = 130
        Inequality gamma < alpha + beta is not satisfied with numerical tolerance: 0.0001

    """
    parallelepiped_check(a, b, c, alpha, beta, gamma, raise_error=True)
    alpha = alpha * TORADIANS
    beta = beta * TORADIANS
    gamma = gamma * TORADIANS
    return np.array(
        [
            [a, 0, 0],
            [b * cos(gamma), b * sin(gamma), 0],
            [
                c * cos(beta),
                c / sin(gamma) * (cos(alpha) - cos(beta) * cos(gamma)),
                c
                / sin(gamma)
                * sqrt(
                    1
                    + 2 * cos(alpha) * cos(beta) * cos(gamma)
                    - cos(alpha) ** 2
                    - cos(beta) ** 2
                    - cos(gamma) ** 2
                ),
            ],
        ],
        dtype=float,
    )


def get_params(cell):
    r"""
    Computes lattice parameters of the cell.

    Parameters
    ----------
    cell : (3, 3) |array-like|_
        Matrix of a cell, rows are interpreted as vectors.

    Returns
    -------
    a : float
        Length of the :math:`\boldsymbol{a_1}` vector.
    b : float
        Length of the :math:`\boldsymbol{a_2}` vector.
    c : float
        Length of the :math:`\boldsymbol{a_3}` vector.
    alpha : float
        Angle between vectors :math:`\boldsymbol{a_2}` and :math:`\boldsymbol{a_3}`
        in degrees.
    beta : float
        Angle between vectors :math:`\boldsymbol{a_1}` and :math:`\boldsymbol{a_3}`
        in degrees.
    gamma : float
        Angle between vectors :math:`\boldsymbol{a_1}` and :math:`\boldsymbol{a_2}`
        in degrees.

    See Also
    --------
    from_params

    Examples
    --------

    .. doctest::

        >>> import wulfric as wulf
        >>> cell = [[1,0,0],[0,2,0],[0,0,3]]
        >>> wulf.cell.get_params(cell)
        (1.0, 2.0, 3.0, 90.0, 90.0, 90.0)
    """

    return (
        float(np.linalg.norm(cell[0])),
        float(np.linalg.norm(cell[1])),
        float(np.linalg.norm(cell[2])),
        get_angle(cell[1], cell[2]),
        get_angle(cell[0], cell[2]),
        get_angle(cell[0], cell[1]),
    )


def get_scalar_products(cell):
    r"""
    Returns pairwise scalar products of the lattice vectors:

    Parameters
    ----------
    cell : (3, 3) |array-like|_
        Matrix of a cell, rows are interpreted as vectors.

    Returns
    -------
    sp_23 : float
        Scalar product of the vectors :math:`\boldsymbol{a}_2\cdot\boldsymbol{a}_3`.
    sp_13 : float
        Scalar product of the vectors :math:`\boldsymbol{a}_1\cdot\boldsymbol{a}_3`.
    sp_12 : float
        Scalar product of the vectors :math:`\boldsymbol{a}_1\cdot\boldsymbol{a}_2`.

    Examples
    --------

    .. doctest::

        >>> import wulfric as wulf
        >>> cell = [[1, 0, 0], [0, 1, 0], [0, 0, 1]]
        >>> wulf.cell.get_scalar_products(cell)
        (0.0, 0.0, 0.0)
    """

    return (
        float(np.dot(cell[1], cell[2])),
        float(np.dot(cell[0], cell[2])),
        float(np.dot(cell[0], cell[1])),
    )


def is_reasonable(cell, eps_lengths=1e-10, eps_volume=1e-5):
    r"""
    Check if the cell is *reasonable* (not *degenerate*) as defined in [1]_.

    The sell is *degenerate* if

    (i)   The minimum of the cell lengths divided by the maximum of the cell lengths is
          smaller than a certain factor :math:`\varepsilon_{lengths}`.
    (ii)  The unit-cell volume divided by the minimum of the cell
          lengths is smaller than a certain factor :math:`\varepsilon_{volume}`.

    The cell is *reasonable* if it is *not degenerate*.

    Parameters
    ----------
    cell : (3, 3) |array-like|_
        Matrix of a cell, rows are interpreted as vectors.
    eps_lengths : float, default 1e-10
        Default value of :math:`\varepsilon_{lengths}`.
    eps_volume : float, default 1e-5
        Default value of :math:`\varepsilon_{volume}`.

    Returns
    -------
    reasonable : bool
        ``True`` if the cell is *reasonable*, ``False`` if cell is *degenerate*.

    References
    ----------
    .. [1] Grosse-Kunstleve, R.W., Sauter, N.K. and Adams, P.D., 2004.
        Numerically stable algorithms for the computation of reduced unit cells.
        Acta Crystallographica Section A: Foundations of Crystallography,
        60(1), pp.1-6.
    """

    cell_volume = get_volume(cell)

    # To guarantee finite max and min lengths
    if cell_volume == 0.0:
        return False

    min_length = np.linalg.norm(cell, axis=1).min()
    max_length = np.linalg.norm(cell, axis=1).max()

    if max_length == 0.0 or min_length / max_length < eps_lengths:
        return False

    if min_length == 0.0 or cell_volume / min_length < eps_volume:
        return False

    return True


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# Wulfric - Cell, Atoms, K-path.
# Copyright (C) 2023-2025 Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


class AbstractBackend:
    def __init__(self) -> None:
        self.kinds = {
            "conventional": self.plot_conventional,
            "primitive": self.plot_primitive,
            "brillouin": self.plot_brillouin,
            "kpath": self.plot_kpath,
            "brillouin-kpath": self.plot_brillouin_kpath,
            "brillouin_kpath": self.plot_brillouin_kpath,
            "wigner-seitz": self.plot_wigner_seitz,
            "wigner_seitz": self.plot_wigner_seitz,
            "unit-cell": self.plot_unit_cell,
            "unit_cell": self.plot_unit_cell,
        }

    # Backend-independent functions
    def plot(self, *args, kind, **kwargs):
        r"""
        Main plotting method.

        Actual list of supported kinds can be check with:

        .. doctest::

            >>> self.kinds.keys() # doctest: +SKIP

        Parameters
        ----------
        kind : str or list of str
            Type of the plot to be plotted. Supported plots:

            * "conventional"
            * "primitive"
            * "brillouin"
            * "kpath"
            * "brillouin-kpath"
            * "wigner-seitz"
            * "unit-cell"
        *args
            Passed directly to the plotting functions.
        **kwargs
            Passed directly to the plotting functions.

        Raises
        ------
        ValueError
            If the plot kind is not supported.

        See Also
        --------
        plot_conventional : "conventional" plot.
        plot_primitive : "primitive" plot.
        plot_brillouin : "brillouin" plot.
        plot_kpath : "kpath" plot.
        plot_brillouin_kpath : "brillouin_kpath" plot.
        plot_wigner_seitz : "wigner-seitz" plot.
        plot_unit_cell : "unit-cell" plot.
        show : Shows the plot.
        save : Save the figure in the file.
        """
        if isinstance(kind, str):
            kinds = [kind]
        else:
            kinds = kind
        for kind in kinds:
            if kind in self.kinds:
                self.kinds[kind](*args, **kwargs)
            else:
                raise ValueError(f"Plot kind '{kind}' does not exist!")

    # Backend-dependent functions
    def remove(self, *args, **kwargs):
        raise NotImplementedError

    def show(self, *args, **kwargs):
        raise NotImplementedError

    def save(self, *args, **kwargs):
        raise NotImplementedError

    def clear(self, *args, **kwargs):
        raise NotImplementedError

    def legend(self, *args, **kwargs):
        raise NotImplementedError

    # Backend-independent functions
    def plot_brillouin(self, *args, color="#FF4D67", **kwargs):
        r"""
        Plot brillouin zone.

        Parameters
        ----------
        *args
            Passed to the :py:meth:`.plot_wigner_seitz` function.
        color : str, default "#FF4D67"
            Colour for the brillouin zone. Any format supported by the used backend.
        **kwargs
            Passed to the :py:meth:`.plot_wigner_seitz` function.

        See Also
        --------
        plot_wigner_seitz : for the list of parameters
        """

        self.plot_wigner_seitz(*args, reciprocal=True, color="#FF4D67", **kwargs)

    def plot_brillouin_kpath(
        self, *args, zone_color="#FF4D67", path_color="black", **kwargs
    ):
        r"""
        Plot brillouin zone and kpath.

        Parameters
        ----------
        *args
            Passed to the :py:meth:`.plot_brillouin` and :py:meth:`.plot_kpath` functions.
        zone_color : str, default "#FF4D67"
            Colour for the brillouin zone. Any format supported by the used backend.
        zone_color : str, default "black"
            Colour for the k path. Any format supported by the used backend.
        **kwargs
            Passed to the :py:meth:`.plot_brillouin` and :py:meth:`.plot_kpath` functions.

        See Also
        --------
        plot_brillouin : plot brillouin zone
        plot_kpath : plot kpath
        """

        self.plot_brillouin(*args, color=zone_color, **kwargs)
        self.plot_kpath(*args, color=path_color, **kwargs)

    def plot_primitive(self, *args, **kwargs):
        r"""
        Plot primitive unit cell.

        Parameters
        ----------
        **kwargs
            Passed to the :py:meth:`.plot_unit_cell` function.

        See Also
        --------
        plot_unit_cell : for the list of parameters
        """

        self.plot_unit_cell(*args, conventional=False, **kwargs)

    def plot_conventional(self, *args, **kwargs):
        r"""
        Plot conventional unit cell.

        See Also
        --------
        plot_unit_cell : for the list of parameters
        """

        self.plot_unit_cell(*args, conventional=True, **kwargs)

    # Backend-dependent functions
    def plot_unit_cell(self, *args, **kwargs):
        raise NotImplementedError

    def plot_wigner_seitz(self, *args, **kwargs):
        raise NotImplementedError

    def plot_kpath(self, *args, **kwargs):
        raise NotImplementedError



# Wulfric - Cell, Atoms, K-path.
# Copyright (C) 2023-2025 Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import numpy as np

from wulfric._exceptions import StandardizationTypeMismatch
from wulfric._numerical import compare_numerically
from wulfric.cell._basic_manipulation import (
    get_params,
    get_reciprocal,
    get_scalar_products,
)
from wulfric.cell._lepage import lepage
from wulfric.constants._sc_notation import C_MATRICES

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


def _CUB_get_S_matrix(cell, length_tolerance=1e-8, angle_tolerance=1e-4):
    r"""
    For arbitrary cubic cell returns matrix S that transforms it to the standardized form.

    See :ref:`guide_cub` and :ref:`user-guide_conventions_which-cell_standardization` for
    details.

    Parameters
    ----------
    cell : (3, 3) |array-like|_
        Matrix of a primitive cell, rows are interpreted as vectors.
    length_tolerance : float, default :math:`10^{-8}`
        Tolerance for length variables (lengths of the lattice vectors). Completely
        ignored by this function, the arguments are defined only for the homogeneity of
        the input for all 14 Bravais lattice types.
    angle_tolerance : float, default :math:`10^{-4}`
        Tolerance for angle variables (angles of the lattice). Completely ignored by this
        function, the arguments are defined only for the homogeneity of the input for all
        14 Bravais lattice types.

    Returns
    -------
    S : (3, 3) :numpy:`ndarray`
        Transformation matrix :math:`S`.

    Notes
    -----
    It is assumed that the ``cell`` has the symmetries of the cubic lattice.
    If the cell is not cubic, the function will not work correctly.
    """

    return np.eye(3, dtype=float)


def _FCC_get_S_matrix(cell, length_tolerance=1e-8, angle_tolerance=1e-4):
    r"""
    For arbitrary face-centered cubic cell returns matrix S that transforms it to the
    standardized form.

    See :ref:`guide_fcc` and :ref:`user-guide_conventions_which-cell_standardization` for
    details.

    Parameters
    ----------
    cell : (3, 3) |array-like|_
        Matrix of a primitive cell, rows are interpreted as vectors.
    length_tolerance : float, default :math:`10^{-8}`
        Tolerance for length variables (lengths of the lattice vectors). Completely
        ignored by this function, the arguments are defined only for the homogeneity of
        the input for all 14 Bravais lattice types.
    angle_tolerance : float, default :math:`10^{-4}`
        Tolerance for angle variables (angles of the lattice). Completely ignored by this
        function, the arguments are defined only for the homogeneity of the input for all
        14 Bravais lattice types.

    Returns
    -------
    S : (3, 3) :numpy:`ndarray`
        Transformation matrix :math:`S`.

    Notes
    -----
    It is assumed that the ``cell`` has the symmetries of the face-centered cubic lattice.
    If the cell is not face-centered cubic, the function will not work correctly.
    """

    return np.eye(3, dtype=float)


def _BCC_get_S_matrix(cell, length_tolerance=1e-8, angle_tolerance=1e-4):
    r"""
    For arbitrary body-centered cubic cell returns matrix S that transforms it to the
    standardized form.

    See :ref:`guide_fcc` and :ref:`user-guide_conventions_which-cell_standardization` for
    details.

    Parameters
    ----------
    cell : (3, 3) |array-like|_
        Matrix of a primitive cell, rows are interpreted as vectors.
    length_tolerance : float, default :math:`10^{-8}`
        Tolerance for length variables (lengths of the lattice vectors). Completely
        ignored by this function, the arguments are defined only for the homogeneity of
        the input for all 14 Bravais lattice types.
    angle_tolerance : float, default :math:`10^{-4}`
        Tolerance for angle variables (angles of the lattice). Completely ignored by this
        function, the arguments are defined only for the homogeneity of the input for all
        14 Bravais lattice types.

    Returns
    -------
    S : (3, 3) :numpy:`ndarray`
        Transformation matrix :math:`S`.

    Notes
    -----
    It is assumed that the ``cell`` has the symmetries of the body-centered cubic lattice.
    If the cell is not body-centered cubic, the function will not work correctly.
    """

    return np.eye(3, dtype=float)


def _TET_get_S_matrix(cell, length_tolerance=1e-8, angle_tolerance=1e-4):
    r"""
    For arbitrary tetragonal cell returns matrix S that transforms it to the
    standardized form.

    See :ref:`guide_tet` and :ref:`user-guide_conventions_which-cell_standardization` for
    details.

    Parameters
    ----------
    cell : (3, 3) |array-like|_
        Matrix of a primitive cell, rows are interpreted as vectors.
    length_tolerance : float, default :math:`10^{-8}`
        Tolerance for length variables (lengths of the lattice vectors). Default value is
        chosen in the contexts of condense matter physics, assuming that length is given
        in Angstroms. Please choose appropriate tolerance for your problem.
    angle_tolerance : float, default :math:`10^{-4}`
        Tolerance for angle variables (angles of the lattice). Completely ignored by this
        function, the arguments are defined only for the homogeneity of the input for all
        14 Bravais lattice types.

    Returns
    -------
    S : (3, 3) :numpy:`ndarray`
        Transformation matrix :math:`S`

    Notes
    -----
    It is assumed that the ``cell`` has the symmetries of the tetragonal lattice. If the
    cell is not tetragonal, the function will not work correctly.

    Raises
    ------
    wulfric.exceptions.StandardizationTypeMismatch
        If none of the tetragonal conditions are satisfied.
    """

    a, b, c, _, _, _ = get_params(cell)

    if compare_numerically(a, "==", b, eps=length_tolerance) and compare_numerically(
        b, "!=", c, eps=length_tolerance
    ):
        S = np.eye(3, dtype=float)
    elif compare_numerically(b, "==", c, eps=length_tolerance) and compare_numerically(
        c, "!=", a, eps=length_tolerance
    ):
        S = np.array([[0, 0, 1], [1, 0, 0], [0, 1, 0]], dtype=float)
    elif compare_numerically(a, "==", c, eps=length_tolerance) and compare_numerically(
        c, "!=", b, eps=length_tolerance
    ):
        S = np.array([[0, 1, 0], [0, 0, 1], [1, 0, 0]], dtype=float)
    else:
        raise StandardizationTypeMismatch("tetragonal")

    return S


def _BCT_get_S_matrix(cell, length_tolerance=1e-8, angle_tolerance=1e-4):
    r"""
    For arbitrary body-centered tetragonal cell returns matrix S that transforms it to the
    standardized form.

    See :ref:`guide_bct` and :ref:`user-guide_conventions_which-cell_standardization` for
    details.

    Parameters
    ----------
    cell : (3, 3) |array-like|_
        Matrix of a primitive cell, rows are interpreted as vectors.
    length_tolerance : float, default :math:`10^{-8}`
        Tolerance for length variables (lengths of the lattice vectors). Completely
        ignored by this function, the arguments are defined only for the homogeneity of
        the input for all 14 Bravais lattice types.
    angle_tolerance : float, default :math:`10^{-4}`
        Tolerance for angle variables (angles of the lattice). Default value is chosen in
        the contexts of condense matter physics, assuming that angles are in degrees.
        Please choose appropriate tolerance for your problem.

    Returns
    -------
    S : (3, 3) :numpy:`ndarray`
        Transformation matrix :math:`S`

    Notes
    -----
    It is assumed that the ``cell`` has the symmetries of the body-centered tetragonal
    lattice. If the cell is not body-centered tetragonal, the function will not work
    correctly.

    Raises
    ------
    wulfric.exceptions.StandardizationTypeMismatch
        If none of the body-centered tetragonal conditions are satisfied.
    """
    cell = np.array(cell)

    _, _, _, alpha, beta, gamma = get_params(cell)

    if compare_numerically(
        alpha, "==", beta, eps=angle_tolerance
    ) and compare_numerically(beta, "!=", gamma, eps=angle_tolerance):
        S = np.eye(3, dtype=float)
    elif compare_numerically(
        beta, "==", gamma, eps=angle_tolerance
    ) and compare_numerically(gamma, "!=", alpha, eps=angle_tolerance):
        S = np.array([[0, 0, 1], [1, 0, 0], [0, 1, 0]], dtype=float)
    elif compare_numerically(
        alpha, "==", gamma, eps=angle_tolerance
    ) and compare_numerically(gamma, "!=", beta, eps=angle_tolerance):
        S = np.array([[0, 1, 0], [0, 0, 1], [1, 0, 0]], dtype=float)
    else:
        raise StandardizationTypeMismatch("body-centered tetragonal")

    return S


def _ORC_get_S_matrix(cell, length_tolerance=1e-8, angle_tolerance=1e-4):
    r"""
    For arbitrary orthorhombic cell returns matrix S that transforms it to the
    standardized form.

    See :ref:`guide_orc` and :ref:`user-guide_conventions_which-cell_standardization` for
    details.

    Parameters
    ----------
    cell : (3, 3) |array-like|_
        Matrix of a primitive cell, rows are interpreted as vectors.
    length_tolerance : float, default :math:`10^{-8}`
        Tolerance for length variables (lengths of the lattice vectors). Default value is
        chosen in the contexts of condense matter physics, assuming that length is given
        in Angstroms. Please choose appropriate tolerance for your problem.
    angle_tolerance : float, default :math:`10^{-4}`
        Tolerance for angle variables (angles of the lattice). Completely ignored by this
        function, the arguments are defined only for the homogeneity of the input for all
        14 Bravais lattice types.

    Returns
    -------
    S : (3, 3) :numpy:`ndarray`
        Transformation matrix :math:`S`

    Notes
    -----
    It is assumed that the ``cell`` has the symmetries of the orthorhombic lattice. If
    the cell is not orthorhombic, the function will not work correctly.

    Raises
    ------
    wulfric.exceptions.StandardizationTypeMismatch
        If none of the orthorhombic conditions are satisfied.
    """

    a, b, c, _, _, _ = get_params(cell)

    if compare_numerically(c, ">", b, eps=length_tolerance) and compare_numerically(
        b, ">", a, eps=length_tolerance
    ):
        S = np.eye(3, dtype=float)
    elif compare_numerically(c, ">", a, eps=length_tolerance) and compare_numerically(
        a, ">", b, eps=length_tolerance
    ):
        S = np.array([[0, -1, 0], [-1, 0, 0], [0, 0, -1]], dtype=float)
    elif compare_numerically(b, ">", c, eps=length_tolerance) and compare_numerically(
        c, ">", a, eps=length_tolerance
    ):
        S = np.array([[-1, 0, 0], [0, 0, -1], [0, -1, 0]], dtype=float)
    elif compare_numerically(b, ">", a, eps=length_tolerance) and compare_numerically(
        a, ">", c, eps=length_tolerance
    ):
        S = np.array([[0, 1, 0], [0, 0, 1], [1, 0, 0]], dtype=float)
    elif compare_numerically(a, ">", c, eps=length_tolerance) and compare_numerically(
        c, ">", b, eps=length_tolerance
    ):
        S = np.array([[0, 0, 1], [1, 0, 0], [0, 1, 0]], dtype=float)
    elif compare_numerically(a, ">", b, eps=length_tolerance) and compare_numerically(
        b, ">", c, eps=length_tolerance
    ):
        S = np.array([[0, 0, -1], [0, -1, 0], [-1, 0, 0]], dtype=float)
    else:
        raise StandardizationTypeMismatch("orthorhombic")

    return S


def _ORCF_get_S_matrix(cell, length_tolerance=1e-8, angle_tolerance=1e-4):
    r"""
    For arbitrary face-centered orthorhombic cell returns matrix S that transforms it to
    the standardized form.

    See :ref:`guide_orcf` and :ref:`user-guide_conventions_which-cell_standardization` for
    details.

    Parameters
    ----------
    cell : (3, 3) |array-like|_
        Matrix of a primitive cell, rows are interpreted as vectors.
    length_tolerance : float, default :math:`10^{-8}`
        Tolerance for length variables (lengths of the lattice vectors). Default value is
        chosen in the contexts of condense matter physics, assuming that length is given
        in Angstroms. Please choose appropriate tolerance for your problem.
    angle_tolerance : float, default :math:`10^{-4}`
        Tolerance for angle variables (angles of the lattice). Completely ignored by this
        function, the arguments are defined only for the homogeneity of the input for all
        14 Bravais lattice types.

    Returns
    -------
    S : (3, 3) :numpy:`ndarray`
        Transformation matrix :math:`S`

    Notes
    -----
    It is assumed that the ``cell`` has the symmetries of the face-centered orthorhombic
    lattice. If the cell is not face-centered orthorhombic, the function will not work
    correctly.

    Raises
    ------
    wulfric.exceptions.StandardizationTypeMismatch
        If none of the face-centered orthorhombic conditions are satisfied.
    """

    a, b, c, _, _, _ = get_params(cell)

    if compare_numerically(c, "<", b, eps=length_tolerance) and compare_numerically(
        b, "<", a, eps=length_tolerance
    ):
        S = np.eye(3, dtype=float)
    elif compare_numerically(c, "<", a, eps=length_tolerance) and compare_numerically(
        a, "<", b, eps=length_tolerance
    ):
        S = np.array([[0, -1, 0], [-1, 0, 0], [0, 0, -1]], dtype=float)
    elif compare_numerically(b, "<", c, eps=length_tolerance) and compare_numerically(
        c, "<", a, eps=length_tolerance
    ):
        S = np.array([[-1, 0, 0], [0, 0, -1], [0, -1, 0]], dtype=float)
    elif compare_numerically(b, "<", a, eps=length_tolerance) and compare_numerically(
        a, "<", c, eps=length_tolerance
    ):
        S = np.array([[0, 1, 0], [0, 0, 1], [1, 0, 0]], dtype=float)
    elif compare_numerically(a, "<", c, eps=length_tolerance) and compare_numerically(
        c, "<", b, eps=length_tolerance
    ):
        S = np.array([[0, 0, 1], [1, 0, 0], [0, 1, 0]], dtype=float)
    elif compare_numerically(a, "<", b, eps=length_tolerance) and compare_numerically(
        b, "<", c, eps=length_tolerance
    ):
        S = np.array([[0, 0, -1], [0, -1, 0], [-1, 0, 0]], dtype=float)
    else:
        raise StandardizationTypeMismatch("face-centered orthorhombic")

    return S


def _ORCI_get_S_matrix(cell, length_tolerance=1e-8, angle_tolerance=1e-4):
    r"""
    For arbitrary body-centered orthorhombic cell returns matrix S that transforms it to
    the standardized form.

    See :ref:`guide_orci` and :ref:`user-guide_conventions_which-cell_standardization` for
    details.

    Parameters
    ----------
    cell : (3, 3) |array-like|_
        Matrix of a primitive cell, rows are interpreted as vectors.
    length_tolerance : float, default :math:`10^{-8}`
        Tolerance for length variables (lengths of the lattice vectors). Completely
        ignored by this function, the arguments are defined only for the homogeneity of
        the input for all 14 Bravais lattice types.
    angle_tolerance : float, default :math:`10^{-4}`
        Tolerance for angle variables (angles of the lattice). Default value is chosen in
        the contexts of condense matter physics, assuming that angles are in degrees.
        Please choose appropriate tolerance for your problem.

    Returns
    -------
    S : (3, 3) :numpy:`ndarray`
        Transformation matrix :math:`S`

    Notes
    -----
    It is assumed that the ``cell`` has the symmetries of the body-centered orthorhombic
    lattice. If the cell is not body-centered orthorhombic, the function will not work
    correctly.

    Raises
    ------
    wulfric.exceptions.StandardizationTypeMismatch
        If none of the body-centered orthorhombic conditions are satisfied.
    """

    _, _, _, alpha, beta, gamma = get_params(cell)

    if compare_numerically(
        gamma, "<", beta, eps=angle_tolerance
    ) and compare_numerically(beta, "<", alpha, eps=angle_tolerance):
        S = np.eye(3, dtype=float)
    elif compare_numerically(
        gamma, "<", alpha, eps=angle_tolerance
    ) and compare_numerically(alpha, "<", beta, eps=angle_tolerance):
        S = np.array([[0, -1, 0], [-1, 0, 0], [0, 0, -1]], dtype=float)
    elif compare_numerically(
        beta, "<", gamma, eps=angle_tolerance
    ) and compare_numerically(gamma, "<", alpha, eps=angle_tolerance):
        S = np.array([[-1, 0, 0], [0, 0, -1], [0, -1, 0]], dtype=float)
    elif compare_numerically(
        beta, "<", alpha, eps=angle_tolerance
    ) and compare_numerically(alpha, "<", gamma, eps=angle_tolerance):
        S = np.array([[0, 1, 0], [0, 0, 1], [1, 0, 0]], dtype=float)
    elif compare_numerically(
        alpha, "<", gamma, eps=angle_tolerance
    ) and compare_numerically(gamma, "<", beta, eps=angle_tolerance):
        S = np.array([[0, 0, 1], [1, 0, 0], [0, 1, 0]], dtype=float)
    elif compare_numerically(
        alpha, "<", beta, eps=angle_tolerance
    ) and compare_numerically(beta, "<", gamma, eps=angle_tolerance):
        S = np.array([[0, 0, -1], [0, -1, 0], [-1, 0, 0]], dtype=float)
    else:
        raise StandardizationTypeMismatch("body-centered orthorhombic")

    return S


def _ORCC_get_S_matrix(cell, length_tolerance=1e-8, angle_tolerance=1e-4):
    r"""
    For arbitrary base-centered orthorhombic cell returns matrix S that transforms it to
    the standardized form.

    See :ref:`guide_orcc` and :ref:`user-guide_conventions_which-cell_standardization` for
    details.

    Parameters
    ----------
    cell : (3, 3) |array-like|_
        Matrix of a primitive cell, rows are interpreted as vectors.
    length_tolerance : float, default :math:`10^{-8}`
        Tolerance for length variables (lengths of the lattice vectors). Completely
        ignored by this function, the arguments are defined only for the homogeneity of
        the input for all 14 Bravais lattice types.
    angle_tolerance : float, default :math:`10^{-4}`
        Tolerance for angle variables (angles of the lattice). Default value is chosen in
        the contexts of condense matter physics, assuming that angles are in degrees.
        Please choose appropriate tolerance for your problem.

    Returns
    -------
    S : (3, 3) :numpy:`ndarray`
        Transformation matrix :math:`S`

    Notes
    -----
    It is assumed that the ``cell`` has the symmetries of the base-centered orthorhombic
    lattice. If the cell is not base-centered orthorhombic, the function will not work
    correctly.

    Raises
    ------
    wulfric.exceptions.StandardizationTypeMismatch
        If none of the base-centered orthorhombic conditions are satisfied.
    """

    _, _, _, alpha, beta, gamma = get_params(cell)

    if (
        compare_numerically(alpha, "==", 90.0, eps=angle_tolerance)
        and compare_numerically(beta, "==", 90.0, eps=angle_tolerance)
        and compare_numerically(gamma, ">", 90.0, eps=angle_tolerance)
    ):
        S = np.eye(3, dtype=float)
    elif (
        compare_numerically(alpha, "==", 90.0, eps=angle_tolerance)
        and compare_numerically(beta, "==", 90.0, eps=angle_tolerance)
        and compare_numerically(gamma, "<", 90.0, eps=angle_tolerance)
    ):
        S = np.array([[0, -1, 0], [1, 0, 0], [0, 0, 1]], dtype=float)
    elif (
        compare_numerically(beta, "==", 90.0, eps=angle_tolerance)
        and compare_numerically(gamma, "==", 90.0, eps=angle_tolerance)
        and compare_numerically(alpha, ">", 90.0, eps=angle_tolerance)
    ):
        S = np.array([[0, 0, 1], [1, 0, 0], [0, 1, 0]], dtype=float)
    elif (
        compare_numerically(beta, "==", 90.0, eps=angle_tolerance)
        and compare_numerically(gamma, "==", 90.0, eps=angle_tolerance)
        and compare_numerically(alpha, "<", 90.0, eps=angle_tolerance)
    ):
        S = np.array([[0, 0, 1], [0, -1, 0], [1, 0, 0]], dtype=float)

    elif (
        compare_numerically(alpha, "==", 90.0, eps=angle_tolerance)
        and compare_numerically(gamma, "==", 90.0, eps=angle_tolerance)
        and compare_numerically(beta, ">", 90.0, eps=angle_tolerance)
    ):
        S = np.array([[0, 1, 0], [0, 0, 1], [1, 0, 0]], dtype=float)
    elif (
        compare_numerically(alpha, "==", 90.0, eps=angle_tolerance)
        and compare_numerically(gamma, "==", 90.0, eps=angle_tolerance)
        and compare_numerically(beta, "<", 90.0, eps=angle_tolerance)
    ):
        S = np.array([[1, 0, 0], [0, 0, 1], [0, -1, 0]], dtype=float)
    else:
        raise StandardizationTypeMismatch("base-centered orthorhombic")

    return S


def _HEX_get_S_matrix(cell, length_tolerance=1e-8, angle_tolerance=1e-4):
    r"""
    For arbitrary hexagonal cell returns matrix S that transforms it to the standardized
    form.

    See :ref:`guide_hex` and :ref:`user-guide_conventions_which-cell_standardization` for
    details.

    Parameters
    ----------
    cell : (3, 3) |array-like|_
        Matrix of a primitive cell, rows are interpreted as vectors.
    length_tolerance : float, default :math:`10^{-8}`
        Tolerance for length variables (lengths of the lattice vectors). Completely
        ignored by this function, the arguments are defined only for the homogeneity of
        the input for all 14 Bravais lattice types.
    angle_tolerance : float, default :math:`10^{-4}`
        Tolerance for angle variables (angles of the lattice). Default value is chosen in
        the contexts of condense matter physics, assuming that angles are in degrees.
        Please choose appropriate tolerance for your problem.

    Returns
    -------
    S : (3, 3) :numpy:`ndarray`
        Transformation matrix :math:`S`

    Notes
    -----
    It is assumed that the ``cell`` has the symmetries of the hexagonal lattice. If the
    cell is not hexagonal, the function will not work correctly.

    Raises
    ------
    wulfric.exceptions.StandardizationTypeMismatch
        If none of the hexagonal conditions are satisfied.
    """

    # Step 1
    _, _, _, alpha, beta, gamma = get_params(cell)

    if compare_numerically(
        alpha, "==", 90.0, eps=angle_tolerance
    ) and compare_numerically(beta, "==", 90.0, eps=angle_tolerance):
        S1 = np.eye(3, dtype=float)
    elif compare_numerically(
        beta, "==", 90.0, eps=angle_tolerance
    ) and compare_numerically(gamma, "==", 90.0, eps=angle_tolerance):
        S1 = np.array([[0, 0, 1], [1, 0, 0], [0, 1, 0]], dtype=float)
    elif compare_numerically(
        alpha, "==", 90.0, eps=angle_tolerance
    ) and compare_numerically(gamma, "==", 90.0, eps=angle_tolerance):
        S1 = np.array([[0, 1, 0], [0, 0, 1], [1, 0, 0]], dtype=float)
    else:
        raise StandardizationTypeMismatch("hexagonal", step="first")

    # Step 2
    cell1 = S1.T @ cell
    _, _, _, _, _, gamma = get_params(cell)

    if compare_numerically(gamma, "==", 120.0, eps=angle_tolerance):
        S2 = np.eye(3, dtype=float)
    elif compare_numerically(gamma, "==", 60.0, eps=angle_tolerance):
        S2 = np.array([[0, -1, 0], [1, 0, 0], [0, 0, 1]], dtype=float)
    else:
        raise StandardizationTypeMismatch("hexagonal", step="second")

    return S1 @ S2


def _RHL_get_S_matrix(cell, length_tolerance=1e-8, angle_tolerance=1e-4):
    r"""
    For arbitrary rhombohedral cell returns matrix S that transforms it to the standardized
    form.

    See :ref:`guide_rhl` and :ref:`user-guide_conventions_which-cell_standardization` for
    details.

    Parameters
    ----------
    cell : (3, 3) |array-like|_
        Matrix of a primitive cell, rows are interpreted as vectors.
    length_tolerance : float, default :math:`10^{-8}`
        Tolerance for length variables (lengths of the lattice vectors). Completely
        ignored by this function, the arguments are defined only for the homogeneity of
        the input for all 14 Bravais lattice types.
    angle_tolerance : float, default :math:`10^{-4}`
        Tolerance for angle variables (angles of the lattice). Completely ignored by this
        function, the arguments are defined only for the homogeneity of the input for all
        14 Bravais lattice types.

    Returns
    -------
    S : (3, 3) :numpy:`ndarray`
        Transformation matrix :math:`S`

    Notes
    -----
    It is assumed that the ``cell`` has the symmetries of the rhombohedral lattice. If the
    cell is not rhombohedral, the function will not work correctly.
    """

    return np.eye(3, dtype=float)


def _MCL_get_S_matrix(cell, length_tolerance=1e-8, angle_tolerance=1e-4):
    r"""
    For arbitrary monoclinic cell returns matrix S that transforms it to the standardized
    form.

    See :ref:`guide_mcl` and :ref:`user-guide_conventions_which-cell_standardization` for
    details.

    Parameters
    ----------
    cell : (3, 3) |array-like|_
        Matrix of a primitive cell, rows are interpreted as vectors.
    length_tolerance : float, default :math:`10^{-8}`
        Tolerance for length variables (lengths of the lattice vectors). Default value is
        chosen in the contexts of condense matter physics, assuming that length is given
        in Angstroms. Please choose appropriate tolerance for your problem.
    angle_tolerance : float, default :math:`10^{-4}`
        Tolerance for angle variables (angles of the lattice). Default value is chosen in
        the contexts of condense matter physics, assuming that angles are in degrees.
        Please choose appropriate tolerance for your problem.

    Returns
    -------
    S : (3, 3) :numpy:`ndarray`
        Transformation matrix :math:`S`

    Notes
    -----
    It is assumed that the ``cell`` has the symmetries of the monoclinic lattice. If the
    cell is not monoclinic, the function will not work correctly.

    Raises
    ------
    wulfric.exceptions.StandardizationTypeMismatch
        If none of the monoclinic conditions are satisfied.
    """

    # Step 1
    _, _, _, alpha, beta, gamma = get_params(cell)

    if (
        compare_numerically(beta, "==", 90.0, eps=angle_tolerance)
        and compare_numerically(gamma, "==", 90.0, eps=angle_tolerance)
        and compare_numerically(alpha, "!=", 90.0, eps=angle_tolerance)
    ):
        S1 = np.eye(3, dtype=float)
    elif (
        compare_numerically(alpha, "==", 90.0, eps=angle_tolerance)
        and compare_numerically(gamma, "==", 90.0, eps=angle_tolerance)
        and compare_numerically(beta, "!=", 90.0, eps=angle_tolerance)
    ):
        S1 = np.array([[0, 0, 1], [1, 0, 0], [0, 1, 0]], dtype=float)
    elif (
        compare_numerically(beta, "==", 90.0, eps=angle_tolerance)
        and compare_numerically(alpha, "==", 90.0, eps=angle_tolerance)
        and compare_numerically(gamma, "!=", 90.0, eps=angle_tolerance)
    ):
        S1 = np.array([[0, 1, 0], [0, 0, 1], [1, 0, 0]], dtype=float)
    else:
        raise StandardizationTypeMismatch("monoclinic", step="First")

    # Step 2
    cell1 = S1.T @ cell
    _, b, c, _, _, _ = get_params(cell1)

    if compare_numerically(b, "<=", c, eps=length_tolerance):
        S2 = np.eye(3, dtype=float)
    elif compare_numerically(b, ">", c, eps=length_tolerance):
        S2 = np.array([[-1, 0, 0], [0, 0, 1], [0, 1, 0]], dtype=float)
    else:
        raise StandardizationTypeMismatch("monoclinic", step="Second")

    # Step 3
    cell2 = S2.T @ cell1
    _, _, _, alpha, _, _ = get_params(cell1)

    if compare_numerically(alpha, "<", 90.0, eps=angle_tolerance):
        S3 = np.eye(3, dtype=float)
    elif compare_numerically(alpha, ">", 90.0, eps=angle_tolerance):
        S3 = np.array([[-1, 0, 0], [0, -1, 0], [0, 0, 1]], dtype=float)
    else:
        raise StandardizationTypeMismatch("monoclinic", step="Third")

    return S1 @ S2 @ S3


def _MCLC_get_S_matrix(cell, length_tolerance=1e-8, angle_tolerance=1e-4):
    r"""
    For arbitrary base-centered monoclinic cell returns matrix S that transforms it to the
    standardized form.

    See :ref:`guide_mclc` and :ref:`user-guide_conventions_which-cell_standardization` for
    details.

    Parameters
    ----------
    cell : (3, 3) |array-like|_
        Matrix of a primitive cell, rows are interpreted as vectors.
    length_tolerance : float, default :math:`10^{-8}`
        Tolerance for length variables (lengths of the lattice vectors). Default value is
        chosen in the contexts of condense matter physics, assuming that length is given
        in Angstroms. Please choose appropriate tolerance for your problem.
    angle_tolerance : float, default :math:`10^{-4}`
        Tolerance for angle variables (angles of the lattice). Default value is chosen in
        the contexts of condense matter physics, assuming that angles are in degrees.
        Please choose appropriate tolerance for your problem.

    Returns
    -------
    S : (3, 3) :numpy:`ndarray`
        Transformation matrix :math:`S`

    Notes
    -----
    It is assumed that the ``cell`` has the symmetries of the base-centered monoclinic
    lattice. If the cell is not base-centered monoclinic, the function will not work
    correctly.

    Raises
    ------
    wulfric.exceptions.StandardizationTypeMismatch
        If none of the base-centered monoclinic conditions are satisfied.
    """

    C = get_C_matrix("MCLC")

    # Step 1
    a, b, c, _, _, _ = get_params(cell)

    if compare_numerically(a, "==", b, eps=length_tolerance):
        S1 = np.eye(3, dtype=float)
    elif compare_numerically(b, "==", c, eps=length_tolerance):
        S1 = np.array([[0, 0, 1], [1, 0, 0], [0, 1, 0]], dtype=float)
    elif compare_numerically(c, "==", a, eps=length_tolerance):
        S1 = np.array([[0, 1, 0], [0, 0, 1], [1, 0, 0]], dtype=float)
    else:
        raise StandardizationTypeMismatch("base-centered monoclinic", step="First")

    # Step 2
    cell1 = S1.T @ cell
    cell1_c = C.T @ cell1
    _, b, c, _, _, _ = get_params(cell1_c)

    if compare_numerically(b, "<=", c, eps=length_tolerance):
        S2_c = np.eye(3, dtype=float)
    else:
        S2_c = np.array([[-1, 0, 0], [0, 0, 1], [0, 1, 0]], dtype=float)

    # Step 3
    cell2_c = S2_c.T @ cell1_c
    _, _, _, alpha, _, _ = get_params(cell2_c)

    if compare_numerically(alpha, "<", 90.0, eps=angle_tolerance):
        S3_c = np.eye(3, dtype=float)
    elif compare_numerically(alpha, ">", 90.0, eps=angle_tolerance):
        S3_c = np.array([[-1, 0, 0], [0, -1, 0], [0, 0, 1]], dtype=float)
    else:
        raise StandardizationTypeMismatch("base-centered monoclinic", step="Third")

    return S1 @ C @ S2_c @ S3_c @ np.linalg.inv(C)


def _TRI_get_S_matrix(cell, length_tolerance=1e-8, angle_tolerance=1e-4):
    r"""
    For arbitrary triclinic cell returns matrix S that transforms it to the
    standardized form.

    See :ref:`guide_tri` and :ref:`user-guide_conventions_which-cell_standardization` for
    details.

    Parameters
    ----------
    cell : (3, 3) |array-like|_
        Matrix of a primitive cell, rows are interpreted as vectors.
    length_tolerance : float, default :math:`10^{-8}`
        Tolerance for length variables (lengths of the lattice vectors). Completely
        ignored by this function, the arguments are defined only for the homogeneity of
        the input for all 14 Bravais lattice types.
    angle_tolerance : float, default :math:`10^{-4}`
        Tolerance for angle variables (angles of the lattice). Default value is chosen in
        the contexts of condense matter physics, assuming that angles are in degrees.
        Please choose appropriate tolerance for your problem.

    Returns
    -------
    S : (3, 3) :numpy:`ndarray`
        Transformation matrix :math:`S`

    Notes
    -----
    It is assumed that the ``cell`` has the symmetries of the triclinic lattice. If the
    cell is not triclinic, the function will not work correctly.

    Raises
    ------
    wulfric.exceptions.StandardizationTypeMismatch
        If none of the triclinic conditions are satisfied.
    """

    # Compute reciprocal cell
    rcell = get_reciprocal(cell)

    # Step 1
    _, _, _, alpha, beta, gamma = get_params(rcell)

    if (
        compare_numerically(alpha, ">=", 90.0, eps=angle_tolerance)
        and compare_numerically(beta, ">=", 90.0, eps=angle_tolerance)
        and compare_numerically(gamma, ">=", 90.0, eps=angle_tolerance)
    ) or (
        compare_numerically(alpha, "<=", 90.0, eps=angle_tolerance)
        and compare_numerically(beta, "<=", 90.0, eps=angle_tolerance)
        and compare_numerically(gamma, "<=", 90.0, eps=angle_tolerance)
    ):
        S1 = np.eye(3, dtype=float)
    elif (
        compare_numerically(alpha, ">=", 90.0, eps=angle_tolerance)
        and compare_numerically(beta, ">=", 90.0, eps=angle_tolerance)
        and compare_numerically(gamma, "<=", 90.0, eps=angle_tolerance)
    ) or (
        compare_numerically(alpha, "<=", 90.0, eps=angle_tolerance)
        and compare_numerically(beta, "<=", 90.0, eps=angle_tolerance)
        and compare_numerically(gamma, ">=", 90.0, eps=angle_tolerance)
    ):
        S1 = np.array([[-1, 0, 0], [0, -1, 0], [0, 0, 1]], dtype=float)
    elif (
        compare_numerically(alpha, ">=", 90.0, eps=angle_tolerance)
        and compare_numerically(beta, "<=", 90.0, eps=angle_tolerance)
        and compare_numerically(gamma, ">=", 90.0, eps=angle_tolerance)
    ) or (
        compare_numerically(alpha, "<=", 90.0, eps=angle_tolerance)
        and compare_numerically(beta, ">=", 90.0, eps=angle_tolerance)
        and compare_numerically(gamma, "<=", 90.0, eps=angle_tolerance)
    ):
        S1 = np.array([[-1, 0, 0], [0, 1, 0], [0, 0, -1]], dtype=float)
    elif (
        compare_numerically(alpha, "<=", 90.0, eps=angle_tolerance)
        and compare_numerically(beta, ">=", 90.0, eps=angle_tolerance)
        and compare_numerically(gamma, ">=", 90.0, eps=angle_tolerance)
    ) or (
        compare_numerically(alpha, ">=", 90.0, eps=angle_tolerance)
        and compare_numerically(beta, "<=", 90.0, eps=angle_tolerance)
        and compare_numerically(gamma, "<=", 90.0, eps=angle_tolerance)
    ):
        S1 = np.array([[1, 0, 0], [0, -1, 0], [0, 0, -1]], dtype=float)
    else:
        raise StandardizationTypeMismatch("triclinic", step="First")

    # Step 2
    rcell1 = S1.T @ rcell
    _, _, _, alpha, beta, gamma = get_params(rcell1)

    if (
        gamma == min(alpha, beta, gamma)
        and compare_numerically(gamma, ">=", 90.0, eps=angle_tolerance)
        or (
            gamma == max(alpha, beta, gamma)
            and compare_numerically(gamma, "<=", 90.0, eps=angle_tolerance)
        )
    ):
        S2 = np.eye(3, dtype=float)
    elif (
        beta == min(alpha, beta, gamma)
        and compare_numerically(beta, ">=", 90.0, eps=angle_tolerance)
        or (
            beta == max(alpha, beta, gamma)
            and compare_numerically(beta, "<=", 90.0, eps=angle_tolerance)
        )
    ):
        S2 = np.array([[0, 1, 0], [0, 0, 1], [1, 0, 0]], dtype=float)
    elif (
        alpha == min(alpha, beta, gamma)
        and compare_numerically(alpha, ">=", 90.0, eps=angle_tolerance)
        or (
            alpha == max(alpha, beta, gamma)
            and compare_numerically(alpha, "<=", 90.0, eps=angle_tolerance)
        )
    ):
        S2 = np.array([[0, 0, 1], [1, 0, 0], [0, 1, 0]], dtype=float)
    else:
        raise StandardizationTypeMismatch("triclinic", step="Second")

    return np.linalg.inv(S1 @ S2).T


def get_S_matrix(cell, lattice_type=None, length_tolerance=1e-8, angle_tolerance=1e-4):
    r"""
    Computes standardization matrix :math:`\boldsymbol{S}` as defined in [1]_.

    See :ref:`docs for each Bravais lattice <user-guide_conventions_bravais-lattices>` for
    the details.

    Parameters
    ----------
    cell : (3, 3) |array-like|_
        Matrix of a primitive cell, rows are interpreted as vectors.
    lattice_type : str, optional
        One of the 14 lattice types that correspond to the provided ``cell``,
        case-insensitive. If not provided, then computed automatically from ``cell``. If
        provided, then it user's responsibility to ensure that ``lattice_type`` is
        correct.
    length_tolerance : float, default :math:`10^{-8}`
        Tolerance for length variables (lengths of the lattice vectors).  Default value is
        chosen in the contexts of condense matter physics, assuming that length is given
        in Angstroms. Please choose appropriate tolerance for your problem.
    angle_tolerance : float, default :math:`10^{-4}`
        Tolerance for angle variables (angles of the lattice). Default value is chosen in
        the contexts of condense matter physics, assuming that angles are in degrees.
        Please choose appropriate tolerance for your problem.

    Returns
    -------
    S : (3, 3) :numpy:`ndarray`
        Transformation matrix :math:`S`

    References
    ----------
    .. [1] Setyawan, W. and Curtarolo, S., 2010.
        High-throughput electronic band structure calculations: Challenges and tools.
        Computational materials science, 49(2), pp. 299-312.

    Examples
    --------

    .. doctest::

        >>> import wulfric as wulf
        >>> cell = [[3, 0, 0],[0, 1, 0],[0, 0, 2]]
        >>> wulf.cell.get_S_matrix(cell)
        array([[0., 0., 1.],
               [1., 0., 0.],
               [0., 1., 0.]])
    """
    cell = np.array(cell, dtype=float)

    if lattice_type is None:
        lattice_type = lepage(cell, angle_tolerance=angle_tolerance)

    lattice_type = lattice_type.upper()

    functions = {
        "CUB": _CUB_get_S_matrix,
        "FCC": _FCC_get_S_matrix,
        "BCC": _BCC_get_S_matrix,
        "TET": _TET_get_S_matrix,
        "BCT": _BCT_get_S_matrix,
        "ORC": _ORC_get_S_matrix,
        "ORCF": _ORCF_get_S_matrix,
        "ORCI": _ORCI_get_S_matrix,
        "ORCC": _ORCC_get_S_matrix,
        "HEX": _HEX_get_S_matrix,
        "RHL": _RHL_get_S_matrix,
        "MCL": _MCL_get_S_matrix,
        "MCLC": _MCLC_get_S_matrix,
        "TRI": _TRI_get_S_matrix,
    }

    return functions[lattice_type](
        cell, length_tolerance=length_tolerance, angle_tolerance=angle_tolerance
    )


def get_C_matrix(lattice_type):
    r"""
    Computes transformation matrix from standardized primitive to standardized
    conventional cell :math:`\boldsymbol{S}` as defined in [1]_.

    See :ref:`user-guide_conventions_which-cell_standardization` for details.

    Parameters
    ----------
    lattice_type : str
        One of the 14 lattice types. Case-insensitive.

    Returns
    -------
    C_matrix : (3, 3) :numpy:`ndarray`

    References
    ----------
    .. [1] Setyawan, W. and Curtarolo, S., 2010.
        High-throughput electronic band structure calculations: Challenges and tools.
        Computational materials science, 49(2), pp. 299-312.

    Examples
    --------

    .. doctest::

        >>> import wulfric as wulf
        >>> wulf.cell.get_C_matrix("ORCF")
        array([[-1.,  1.,  1.],
               [ 1., -1.,  1.],
               [ 1.,  1., -1.]])
    """

    return C_MATRICES[lattice_type.upper()]


def get_standardized(cell, S_matrix=None, length_tolerance=1e-8, angle_tolerance=1e-4):
    R"""
    Computes standardizes primitive cell as defined in [1]_.

    Parameters
    ----------
    cell : (3, 3) |array-like|_
        Matrix of a primitive cell, rows are interpreted as vectors.
    S_matrix : (3, 3) |array-like|_, optional
        Transformation matrix S. If not provided, then computed automatically from
        ``cell``. If provided, then it is user's responsibility to ensure that the matrix
        is the correct one for the given ``cell``.
    length_tolerance : float, default :math:`10^{-8}`
        Tolerance for length variables (lengths of the lattice vectors).  Default value is
        chosen in the contexts of condense matter physics, assuming that length is given
        in Angstroms. Please choose appropriate tolerance for your problem.
    angle_tolerance : float, default :math:`10^{-4}`
        Tolerance for angle variables (angles of the lattice). Default value is chosen in
        the contexts of condense matter physics, assuming that angles are in degrees.
        Please choose appropriate tolerance for your problem.

    Returns
    -------
    cell : (3, 3) :numpy:`ndarray`
        Standardized cell. Rows are interpreted as vectors. Independent from the initial
        cell, safe to modify.

        .. code-block:: python

            cell = [[a1_x, a1_y, a1_z],
                    [a2_x, a2_y, a2_z],
                    [a3_x, a3_y, a3_z]]

    References
    ----------
    .. [1] Setyawan, W. and Curtarolo, S., 2010.
        High-throughput electronic band structure calculations: Challenges and tools.
        Computational materials science, 49(2), pp. 299-312.

    Examples
    --------

    .. doctest::

        >>> import wulfric as wulf
        >>> cell = [[3, 0, 0],[0, 1, 0],[0, 0, 2]]
        >>> wulf.cell.get_standardized(cell)
        array([[0., 1., 0.],
               [0., 0., 2.],
               [3., 0., 0.]])
    """

    cell = np.array(cell, dtype=float)

    if S_matrix is None:
        lattice_type = lepage(cell, angle_tolerance=angle_tolerance)

        S_matrix = get_S_matrix(
            cell,
            lattice_type,
            length_tolerance=length_tolerance,
            angle_tolerance=angle_tolerance,
        )
    else:
        S_matrix = np.array(S_matrix, dtype=float)

    return S_matrix.T @ cell


def get_conventional(
    cell, S_matrix=None, C_matrix=None, length_tolerance=1e-8, angle_tolerance=1e-4
):
    r"""
    Computes standardizes conventional cell as defined in [1]_.

    Parameters
    ----------
    cell : (3, 3) |array-like|_
        Matrix of a primitive cell, rows are interpreted as vectors.
    S_matrix : (3, 3) |array-like|_, optional
        Transformation matrix S. If not provided, then computed automatically from
        ``cell``. If provided, then it is user's responsibility to ensure that the matrix
        is the correct one for the given ``cell``.
    C_matrix : (3, 3) |array-like|_, optional
        Transformation matrix C. If not provided, then computed automatically from
        ``cell``. If provided, then it is user's responsibility to ensure that the matrix
        is the correct one for the given ``cell``.
    length_tolerance : float, default :math:`10^{-8}`
        Tolerance for length variables (lengths of the lattice vectors).  Default value is
        chosen in the contexts of condense matter physics, assuming that length is given
        in Angstroms. Please choose appropriate tolerance for your problem.
    angle_tolerance : float, default :math:`10^{-4}`
        Tolerance for angle variables (angles of the lattice). Default value is chosen in
        the contexts of condense matter physics, assuming that angles are in degrees.
        Please choose appropriate tolerance for your problem.

    Returns
    -------
    conv_cell : (3, 3) :numpy:`ndarray`
        Conventional cell, rows are interpreted as vectors, columns - as coordinates.
        Independent from the initial cell, safe to modify.

        .. code-block:: python

            conv_cell = [[a1_x, a1_y, a1_z],
                         [a2_x, a2_y, a2_z],
                         [a3_x, a3_y, a3_z]]

    References
    ----------
    .. [1] Setyawan, W. and Curtarolo, S., 2010.
        High-throughput electronic band structure calculations: Challenges and tools.
        Computational materials science, 49(2), pp. 299-312.

    Examples
    --------

    .. doctest::

        >>> import wulfric as wulf
        >>> orcf = wulf.cell.ORCF(a=1,b=2,c=3)
        >>> wulf.cell.get_conventional(orcf)
        array([[1., 0., 0.],
               [0., 2., 0.],
               [0., 0., 3.]])
    """
    cell = np.array(cell, dtype=float)

    if S_matrix is None or C_matrix is None:
        lattice_type = lepage(cell, angle_tolerance=angle_tolerance)

    if C_matrix is None:
        C_matrix = get_C_matrix(lattice_type)
    else:
        C_matrix = np.array(C_matrix, dtype=float)

    if S_matrix is None:
        S_matrix = get_S_matrix(
            cell,
            lattice_type,
            length_tolerance=length_tolerance,
            angle_tolerance=angle_tolerance,
        )
    else:
        S_matrix = np.array(S_matrix, dtype=float)

    return C_matrix.T @ S_matrix.T @ cell


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# Wulfric - Cell, Atoms, K-path.
# Copyright (C) 2023-2025 Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from wulfric.constants._sc_notation import BRAVAIS_LATTICE_NAMES, PEARSON_SYMBOLS

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


def get_name(lattice_type):
    r"""
    Human-readable name of the Bravais lattice type.

    Parameters
    ----------
    lattice_type : str, optional
        One of the 14 lattice types. Case-insensitive.

    Returns
    -------
    name : str
        Name of the Bravais lattice type.

    Examples
    --------

    .. doctest::

        >>> import wulfric as wulf
        >>> wulf.cell.get_name("CUB")
        'Cubic'
        >>> wulf.cell.get_name("MCLC")
        'C-centered monoclinic'
    """

    return BRAVAIS_LATTICE_NAMES[lattice_type.upper()]


def get_pearson_symbol(lattice_type):
    r"""
    Pearson symbol.

    Parameters
    ----------
    lattice_type : str, optional
        One of the 14 lattice types. Case-insensitive.

    Returns
    -------
    pearson_symbol : str
        Pearson symbol of the lattice.

    Notes
    -----
    See: |PearsonSymbol|_

    Examples
    --------

    .. doctest::

        >>> import wulfric as wulf
        >>> wulf.cell.get_pearson_symbol("CUB")
        'cP'
        >>> wulf.cell.get_pearson_symbol("MCLC")
        'mS'
        >>> wulf.cell.get_pearson_symbol("Tri")
        'aP'
        >>> wulf.cell.get_pearson_symbol("ORCC")
        'oS'
    """

    return PEARSON_SYMBOLS[lattice_type.upper()]


def get_crystal_family(lattice_type):
    r"""
    Crystal family.

    Parameters
    ----------
    lattice_type : str, optional
        One of the 14 lattice types. Case-insensitive.

    Returns
    -------
    crystal_family : str
        Crystal family of the lattice.

    Notes
    -----
    See: |PearsonSymbol|_

    Examples
    --------

    .. doctest::

        >>> import wulfric as wulf
        >>> wulf.cell.get_crystal_family("CUB")
        'c'
        >>> wulf.cell.get_crystal_family("MCLC")
        'm'
        >>> wulf.cell.get_crystal_family("Tri")
        'a'
        >>> wulf.cell.get_crystal_family("ORCC")
        'o'
    """

    return PEARSON_SYMBOLS[lattice_type.upper()][0]


def get_centring_type(lattice_type):
    r"""
    Centring type.

    Parameters
    ----------
    lattice_type : str, optional
        One of the 14 lattice types. Case-insensitive.

    Returns
    -------
    centring_type : str
        Centring type of the lattice.

    Notes
    -----
    See: |PearsonSymbol|_



    Examples
    --------

    .. doctest::

        >>> import wulfric as wulf
        >>> wulf.cell.get_centring_type("CUB")
        'P'
        >>> wulf.cell.get_centring_type("MCLC")
        'S'
        >>> wulf.cell.get_centring_type("Tri")
        'P'
        >>> wulf.cell.get_centring_type("ORCC")
        'S'
    """

    return PEARSON_SYMBOLS[lattice_type.upper()][1]


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# Wulfric - Cell, Atoms, K-path.
# Copyright (C) 2023-2025 Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import numpy as np
from termcolor import cprint

from wulfric._exceptions import NiggliReductionFailed
from wulfric._numerical import compare_numerically
from wulfric.cell._basic_manipulation import get_params
from wulfric.constants._numerical import TODEGREES
from wulfric.geometry._geometry import get_volume

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


def _niggli_step_1(A, B, C, xi, eta, zeta, trans_matrix, eps):
    condition = compare_numerically(A, ">", B, eps=eps) or (
        compare_numerically(A, "==", B, eps=eps)
        and compare_numerically(abs(xi), ">", abs(eta), eps=eps)
    )
    if condition:

        trans_matrix = trans_matrix @ np.array(
            [
                [0, -1, 0],
                [-1, 0, 0],
                [0, 0, -1],
            ],
            dtype=int,
        )

        A, xi, B, eta = B, eta, A, xi

    return condition, (A, B, C, xi, eta, zeta), trans_matrix


def _niggli_step_2(A, B, C, xi, eta, zeta, trans_matrix, eps):
    condition = compare_numerically(B, ">", C, eps=eps) or (
        compare_numerically(B, "==", C, eps=eps)
        and compare_numerically(abs(eta), ">", abs(zeta), eps=eps)
    )
    if condition:

        trans_matrix = trans_matrix @ np.array(
            [
                [-1, 0, 0],
                [0, 0, -1],
                [0, -1, 0],
            ],
            dtype=int,
        )

        B, eta, C, zeta = C, zeta, B, eta

    return condition, (A, B, C, xi, eta, zeta), trans_matrix


def _niggli_step_3(A, B, C, xi, eta, zeta, trans_matrix, eps):
    condition = compare_numerically(xi * eta * zeta, ">", 0, eps=eps)
    if condition:

        if compare_numerically(xi, ">", 0, eps=eps):
            i = 1
        else:
            i = -1

        if compare_numerically(eta, ">", 0, eps=eps):
            j = 1
        else:
            j = -1

        if compare_numerically(zeta, ">", 0, eps=eps):
            k = 1
        else:
            k = -1

        trans_matrix = trans_matrix @ np.array(
            [
                [i, 0, 0],
                [0, j, 0],
                [0, 0, k],
            ],
            dtype=int,
        )

        xi, eta, zeta = abs(xi), abs(eta), abs(zeta)

    return condition, (A, B, C, xi, eta, zeta), trans_matrix


def _niggli_step_4(A, B, C, xi, eta, zeta, trans_matrix, eps):
    condition = compare_numerically(xi * eta * zeta, "<=", 0, eps=eps)
    if condition:

        # Step 1
        i, j, k = 1, 1, 1
        p = None

        # Step 2
        if compare_numerically(xi, ">", 0):
            i = -1
        elif not compare_numerically(xi, "<", 0):
            p = "i"

        # Step 3
        if compare_numerically(eta, ">", 0):
            j = -1
        elif not compare_numerically(eta, "<", 0):
            p = "j"

        # Step 4
        if compare_numerically(zeta, ">", 0):
            k = -1
        elif not compare_numerically(zeta, "<", 0):
            p = "k"

        # Step 5
        if i * j * k < 0 and p is not None:
            if p == "i":
                i = -1
            elif p == "j":
                j = -1
            elif p == "k":
                k = -1

        trans_matrix = trans_matrix @ np.array(
            [
                [i, 0, 0],
                [0, j, 0],
                [0, 0, k],
            ],
            dtype=int,
        )

        xi, eta, zeta = -abs(xi), -abs(eta), -abs(zeta)

    return condition, (A, B, C, xi, eta, zeta), trans_matrix


def _niggli_step_5(A, B, C, xi, eta, zeta, trans_matrix, eps):
    condition = (
        compare_numerically(abs(xi), ">", B, eps=eps)
        or (
            compare_numerically(xi, "==", B, eps=eps)
            and compare_numerically(2 * eta, "<", zeta, eps=eps)
        )
        or (
            compare_numerically(xi, "==", -B, eps=eps)
            and compare_numerically(zeta, "<", 0, eps=eps)
        )
    )
    if condition:

        trans_matrix = trans_matrix @ np.array(
            [
                [1, 0, 0],
                [0, 1, -np.sign(xi)],
                [0, 0, 1],
            ],
            dtype=int,
        )

        C = B + C - xi * np.sign(xi)
        eta = eta - zeta * np.sign(xi)
        xi = xi - 2 * B * np.sign(xi)

    return condition, (A, B, C, xi, eta, zeta), trans_matrix


def _niggli_step_6(A, B, C, xi, eta, zeta, trans_matrix, eps):
    condition = (
        compare_numerically(abs(eta), ">", A, eps=eps)
        or (
            compare_numerically(eta, "==", A, eps=eps)
            and compare_numerically(2 * xi, "<", zeta, eps=eps)
        )
        or (
            compare_numerically(eta, "==", -A, eps=eps)
            and compare_numerically(zeta, "<", 0, eps=eps)
        )
    )
    if condition:

        trans_matrix = trans_matrix @ np.array(
            [
                [1, 0, -np.sign(eta)],
                [0, 1, 0],
                [0, 0, 1],
            ],
            dtype=int,
        )

        C = A + C - eta * np.sign(eta)
        xi = xi - zeta * np.sign(eta)
        eta = eta - 2 * A * np.sign(eta)

    return condition, (A, B, C, xi, eta, zeta), trans_matrix


def _niggli_step_7(A, B, C, xi, eta, zeta, trans_matrix, eps):
    condition = (
        compare_numerically(abs(zeta), ">", A, eps=eps)
        or (
            compare_numerically(zeta, "==", A, eps=eps)
            and compare_numerically(2 * xi, "<", eta, eps=eps)
        )
        or (
            compare_numerically(zeta, "==", -A, eps=eps)
            and compare_numerically(eta, "<", 0, eps=eps)
        )
    )
    if condition:

        trans_matrix = trans_matrix @ np.array(
            [
                [1, -np.sign(zeta), 0],
                [0, 1, 0],
                [0, 0, 1],
            ],
            dtype=int,
        )

        B = A + B - zeta * np.sign(zeta)
        xi = xi - eta * np.sign(zeta)
        zeta = zeta - 2 * A * np.sign(zeta)

    return condition, (A, B, C, xi, eta, zeta), trans_matrix


def _niggli_step_8(A, B, C, xi, eta, zeta, trans_matrix, eps):
    condition = compare_numerically(xi + eta + zeta + A + B, "<", 0, eps=eps) or (
        compare_numerically(xi + eta + zeta + A + B, "==", 0, eps=eps)
        and compare_numerically(2 * (A + eta) + zeta, ">", 0, eps=eps)
    )
    if condition:

        trans_matrix = trans_matrix @ np.array(
            [
                [1, 0, 1],
                [0, 1, 1],
                [0, 0, 1],
            ],
            dtype=int,
        )

        C = A + B + C + xi + eta + zeta
        xi = 2 * B + xi + zeta
        eta = 2 * A + eta + zeta

    return condition, (A, B, C, xi, eta, zeta), trans_matrix


def niggli(
    cell,
    eps_relative=1e-5,
    max_iterations=100000,
    return_transformation_matrix=False,
):
    r"""
    Computes reduced Niggli cell. Implements algorithm from [2]_.

    Details of the implementation are written in :ref:`library_niggli`.

    Parameters
    ----------
    cell : (3, 3) |array-like|_
        Matrix of a cell, rows are interpreted as vectors.
    eps_relative : float, default :math:`10^{-5}`
        Relative epsilon as defined in [2]_.
    max_iterations : int, default 100000
        Maximum number of iterations.
    return_transformation_matrix : bool, default False
        Whether to return a transformation matrix from given ``cell`` to the
        ``niggli_cell``.

    Returns
    -------
    niggli_cell : (3, 3) :numpy:`ndarray`
        Matrix of a niggli reduced cell, rows are interpreted as vectors.

        .. code-block:: python

            niggli_cell = [[a1_x, a1_y, a1_z],
                           [a2_x, a2_y, a2_z],
                           [a3_x, a3_y, a3_z]]

    transformation_matrix : (3, 3) :numpy:`ndarray`
        Returned only if ``return_transformation_matrix`` is ``True``.

    Raises
    ------
    wulfric.exceptions.NiggliReductionFailed
        If the niggli cell is not found in ``max_iterations`` iterations.
    ValueError
        If the volume of ``cell`` is zero.

    References
    ----------
    .. [1] Křivý, I. and Gruber, B., 1976.
        A unified algorithm for determining the reduced (Niggli) cell.
        Acta Crystallographica Section A: Crystal Physics, Diffraction,
        Theoretical and General Crystallography,
        32(2), pp.297-298.
    .. [2] Grosse-Kunstleve, R.W., Sauter, N.K. and Adams, P.D., 2004.
        Numerically stable algorithms for the computation of reduced unit cells.
        Acta Crystallographica Section A: Foundations of Crystallography,
        60(1), pp.1-6.

    Examples
    --------

    .. doctest::

        >>> import wulfric as wulf
        >>> wulf.cell.niggli([[1, -0.5, 0],[-0.5, 1, 0],[0, 0, 1]])
        array([[ 0.5,  0.5,  0. ],
               [ 0. ,  0. , -1. ],
               [-1. ,  0.5,  0. ]])

    Example from [1]_ (parameters are reproducing :math:`A=9`, :math:`B=27`, :math:`C=4`,
    :math:`\xi` = -5, :math:`\eta` = -4, :math:`\zeta = -22`):

    .. doctest::

        >>> import wulfric as wulf
        >>> from wulfric.constants import TODEGREES
        >>> from math import sqrt, acos
        >>> a = 3
        >>> b = sqrt(27)
        >>> c = 2
        >>> alpha = acos(-5 / 2 / b / c) * TODEGREES
        >>> beta = acos(-4 / 2 / a / c) * TODEGREES
        >>> gamma = acos(-22 / 2 / a / b) * TODEGREES
        >>> cell = wulf.cell.from_params(a, b, c, alpha, beta, gamma)
        >>> niggli_cell = wulf.cell.niggli(cell)
        >>> niggli_cell @ niggli_cell.T
        array([[4. , 2. , 1.5],
               [2. , 9. , 4.5],
               [1.5, 4.5, 9. ]])

    """

    volume = get_volume(cell)
    if volume == 0:
        raise ValueError("Cell volume is zero")

    eps = eps_relative * volume ** (1 / 3.0)

    # 0
    metric_tensor = np.matmul(cell, np.transpose(cell))

    params = (
        metric_tensor[0][0],
        metric_tensor[1][1],
        metric_tensor[2][2],
        2 * metric_tensor[1][2],
        2 * metric_tensor[0][2],
        2 * metric_tensor[0][1],
    )

    trans_matrix = np.eye(3, dtype=int)

    iter_count = 0
    while True:

        if iter_count > max_iterations:
            raise NiggliReductionFailed(max_iterations=max_iterations)

        # Note : each iteration changes the transformation matrix

        iter_count += 1
        # 1
        condition, params, trans_matrix = _niggli_step_1(*params, trans_matrix, eps=eps)

        # 2
        condition, params, trans_matrix = _niggli_step_2(*params, trans_matrix, eps=eps)
        if condition:
            continue

        # 3
        condition, params, trans_matrix = _niggli_step_3(*params, trans_matrix, eps=eps)

        # 4
        condition, params, trans_matrix = _niggli_step_4(*params, trans_matrix, eps=eps)

        # 5
        condition, params, trans_matrix = _niggli_step_5(*params, trans_matrix, eps=eps)
        if condition:
            continue

        # 6
        condition, params, trans_matrix = _niggli_step_6(*params, trans_matrix, eps=eps)
        if condition:
            continue

        # 7
        condition, params, trans_matrix = _niggli_step_7(*params, trans_matrix, eps=eps)
        if condition:
            continue

        # 8
        condition, params, trans_matrix = _niggli_step_8(*params, trans_matrix, eps=eps)
        if condition:
            continue

        break

    niggli_cell = trans_matrix.T @ cell

    if return_transformation_matrix:
        return niggli_cell, trans_matrix

    return niggli_cell


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# Wulfric - Cell, Atoms, K-path.
# Copyright (C) 2023-2025 Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from random import choices
from string import ascii_lowercase
from typing import Iterable

import numpy as np

from wulfric._kpoints_class import Kpoints
from wulfric.cell._basic_manipulation import get_reciprocal
from wulfric.cell._sc_standardize import get_conventional
from wulfric.cell._voronoi import _get_voronoi_cell
from wulfric.geometry._geometry import get_volume
from wulfric.visualization._interface import AbstractBackend

try:
    import plotly.graph_objects as go

    PLOTLY_AVAILABLE = True
except ImportError:
    PLOTLY_AVAILABLE = False

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


class PlotlyBackend(AbstractBackend):
    r"""
    Plotting engine based on |plotly|_.

    Parameters
    ----------
    fig : plotly graph object
        Figure to plot on. If not provided, a new figure is created.

    Attributes
    ----------
    fig : plotly graph object
        Figure to plot on.

    Notes
    -----
    This class is a part of ``wulfric[visual]``
    """

    def __init__(self, fig=None):
        if not PLOTLY_AVAILABLE:
            raise ImportError(
                'Plotly is not available. Install it with "pip install plotly"'
            )
        super().__init__()
        if fig is None:
            fig = go.Figure()
        self.fig = fig

    def show(self, axes_visible=True, **kwargs):
        r"""
        Shows the figure in the interactive mode.

        Parameters
        ----------
        axes_visible : bool, default True
            Whether to show axes.
        **kwargs
            Passed directly to the |plotly-update-layout|_.
        """

        if not axes_visible:
            self.fig.update_scenes(
                xaxis_visible=False, yaxis_visible=False, zaxis_visible=False
            )

        # Set up defaults
        if "width" not in kwargs:
            kwargs["width"] = 800
        if "height" not in kwargs:
            kwargs["height"] = 700
        if "yaxis_scaleanchor" not in kwargs:
            kwargs["yaxis_scaleanchor"] = "x"
        if "showlegend" not in kwargs:
            kwargs["showlegend"] = False
        if "autosize" not in kwargs:
            kwargs["autosize"] = False

        self.fig.update_layout(**kwargs)
        self.fig.show()

    def save(
        self,
        output_name="lattice_graph.png",
        kwargs_update_layout=None,
        kwargs_write_html=None,
        axes_visible=True,
    ):
        r"""
        Saves the figure in the html file.

        Parameters
        ----------
        output_name : str, default "lattice_graph.png"
            Name of the file to be saved. With extension.
        kwargs_update_layout : dict, optional
            Passed directly to the |plotly-update-layout|_.
        kwargs_write_html : dict, optional
            Passed directly to the |plotly-write-html|_.
        axes_visible : bool, default True
            Whether to show axes.
        """

        if kwargs_update_layout is None:
            kwargs_update_layout = {}
        if kwargs_write_html is None:
            kwargs_write_html = {}

        self.fig.update_scenes(aspectmode="data")
        if not axes_visible:
            self.fig.update_scenes(
                xaxis_visible=False, yaxis_visible=False, zaxis_visible=False
            )

        self.fig.update_layout(**kwargs_update_layout)

        self.fig.write_html(output_name, **kwargs_write_html)

    def plot_unit_cell(
        self,
        cell,
        vectors=True,
        color="#274DD1",
        label=None,
        conventional=False,
        reciprocal=False,
        normalize=False,
    ):
        r"""
        Plots real or reciprocal unit cell.

        Parameters
        ----------
        cell : (3, 3) |array-like|_
            Matrix of a cell, rows are interpreted as vectors.
        vectors : bool, default True
            Whether to plot lattice vectors.
        color : str, default "#274DD1"
            Colour for the plot. Any value supported Plotly.
        label : str, optional
            Label for the plot.
        conventional : bool, default False
            Whether to plot conventional cell.
            Only primitive unit cell is supported for reciprocal space.
        reciprocal : bool, default False
            Whether to plot reciprocal or real unit cell.
        normalize : bool, default False
            Whether to normalize volume of the cell to one.
        """
        if reciprocal and conventional:
            raise ValueError("Conventional cell is not supported in reciprocal space.")
        if conventional:
            artist_group = "conventional"
        else:
            artist_group = "primitive"

        if reciprocal:
            artist_group += "_reciprocal"
            vector_label = "b"
        else:
            artist_group += "_real"
            vector_label = "a"

        if conventional:
            cell = get_conventional(cell)
        elif reciprocal:
            cell = get_reciprocal(cell)

        if normalize:
            cell /= abs(get_volume(cell) ** (1 / 3.0))

        legendgroup = "".join(choices(ascii_lowercase, k=10))

        if vectors:
            labels = [f"{vector_label}{i+1}" for i in range(3)]
            for i in range(3):
                x = [0, cell[i][0]]
                y = [0, cell[i][1]]
                z = [0, cell[i][2]]
                self.fig.add_traces(
                    data=[
                        {
                            "x": x,
                            "y": y,
                            "z": z,
                            "mode": "lines",
                            "type": "scatter3d",
                            "hoverinfo": "none",
                            "line": {"color": color, "width": 3},
                            "showlegend": False,
                            "legendgroup": legendgroup,
                        },
                        {
                            "type": "cone",
                            "x": [x[1]],
                            "y": [y[1]],
                            "z": [z[1]],
                            "u": [0.2 * (x[1] - x[0])],
                            "v": [0.2 * (y[1] - y[0])],
                            "w": [0.2 * (z[1] - z[0])],
                            "anchor": "tip",
                            "hoverinfo": "none",
                            "colorscale": [[0, color], [1, color]],
                            "showscale": False,
                            "showlegend": False,
                            "legendgroup": legendgroup,
                        },
                    ]
                )
                self.fig.add_traces(
                    data=go.Scatter3d(
                        mode="text",
                        x=[1.2 * x[1]],
                        y=[1.2 * y[1]],
                        z=[1.2 * z[1]],
                        marker=dict(size=0, color=color),
                        text=labels[i],
                        hoverinfo="none",
                        textposition="top center",
                        textfont=dict(size=12),
                        showlegend=False,
                        legendgroup=legendgroup,
                    )
                )

        def plot_line(line, shift, showlegend=False):
            self.fig.add_traces(
                data=go.Scatter3d(
                    mode="lines",
                    x=[shift[0], shift[0] + line[0]],
                    y=[shift[1], shift[1] + line[1]],
                    z=[shift[2], shift[2] + line[2]],
                    line=dict(color=color),
                    hoverinfo="none",
                    legendgroup=legendgroup,
                    name=label,
                    showlegend=showlegend,
                ),
            )

        showlegend = label is not None
        for i in range(0, 3):
            j = (i + 1) % 3
            k = (i + 2) % 3
            plot_line(cell[i], np.zeros(3), showlegend=showlegend)
            if showlegend:
                showlegend = False
            plot_line(cell[i], cell[j])
            plot_line(cell[i], cell[k])
            plot_line(cell[i], cell[j] + cell[k])

    def plot_wigner_seitz(
        self,
        cell,
        vectors=True,
        label=None,
        color="black",
        reciprocal=False,
        normalize=False,
    ):
        r"""
        Plots Wigner-Seitz cell.

        Parameters
        ----------
        cell : (3, 3) |array-like|_
            Matrix of a cell, rows are interpreted as vectors.
        vectors : bool, default True
            Whether to plot lattice vectors.
        label : str, optional
            Label for the plot.
        color : str, default "black" or "#FF4D67"
            Colour for the plot. Any value supported Plotly.
        reciprocal : bool, default False
            Whether to plot reciprocal or real Wigner-Seitz cell.
        normalize : bool, default False
            Whether to normalize volume of the cell to one.
        """

        if reciprocal:
            cell = get_reciprocal(cell)
            vector_label = "b"
        else:
            vector_label = "a"

        if color is None:
            color = "black"

        if normalize:
            cell /= abs(get_volume(cell) ** (1 / 3.0))

        v1, v2, v3 = cell[0], cell[1], cell[2]

        vs = [v1, v2, v3]

        legendgroup = "".join(choices(ascii_lowercase, k=10))

        if vectors:
            labels = [f"{vector_label}{i+1}" for i in range(3)]
            for i in range(3):
                x = [0, vs[i][0]]
                y = [0, vs[i][1]]
                z = [0, vs[i][2]]
                self.fig.add_traces(
                    data=[
                        {
                            "x": x,
                            "y": y,
                            "z": z,
                            "mode": "lines",
                            "type": "scatter3d",
                            "hoverinfo": "none",
                            "line": {"color": color, "width": 3},
                            "showlegend": False,
                            "legendgroup": legendgroup,
                        },
                        {
                            "type": "cone",
                            "x": [x[1]],
                            "y": [y[1]],
                            "z": [z[1]],
                            "u": [0.2 * (x[1] - x[0])],
                            "v": [0.2 * (y[1] - y[0])],
                            "w": [0.2 * (z[1] - z[0])],
                            "anchor": "tip",
                            "hoverinfo": "none",
                            "colorscale": [[0, color], [1, color]],
                            "showscale": False,
                            "showlegend": False,
                            "legendgroup": legendgroup,
                        },
                    ]
                )
                self.fig.add_traces(
                    data=go.Scatter3d(
                        mode="text",
                        x=[1.2 * x[1]],
                        y=[1.2 * y[1]],
                        z=[1.2 * z[1]],
                        marker=dict(size=0, color=color),
                        text=labels[i],
                        hoverinfo="none",
                        textposition="top center",
                        textfont=dict(size=12),
                        showlegend=False,
                        legendgroup=legendgroup,
                    )
                )

        edges, _ = _get_voronoi_cell(cell)
        showlegend = label is not None
        for p1, p2 in edges:
            xyz = np.array([p1, p2]).T
            self.fig.add_traces(
                data=go.Scatter3d(
                    mode="lines",
                    x=xyz[0],
                    y=xyz[1],
                    z=xyz[2],
                    line=dict(color=color),
                    hoverinfo="none",
                    showlegend=showlegend,
                    legendgroup=legendgroup,
                    name=label,
                ),
            )
            if showlegend:
                showlegend = False

    def plot_kpath(self, cell, color="#000000", label=None, normalize=False, **kwargs):
        r"""
        Plots k path in the reciprocal space.

        Parameters
        ----------
        cell : (3, 3) |array-like|_
            Matrix of a cell, rows are interpreted as vectors.
        color : str, default "#000000"
            Colour for the plot. Any value supported Plotly.
        label : str, optional
            Label for the plot.
        normalize : bool, default False
            Whether to normalize volume of the cell to one.
        """

        if normalize:
            cell /= get_volume(cell) ** (1 / 3.0)

        kp = Kpoints.from_cell(cell=cell)

        cell = get_reciprocal(cell)

        p_abs = []
        p_rel = []
        labels = []
        for point in kp.hs_names:
            p_abs.append(tuple(kp.hs_coordinates[point] @ cell))
            p_rel.append(kp.hs_coordinates[point])

            labels.append(kp.hs_labels[point])

        p_abs = np.array(p_abs).T

        self.fig.add_traces(
            data=go.Scatter3d(
                mode="markers+text",
                x=p_abs[0],
                y=p_abs[1],
                z=p_abs[2],
                marker=dict(size=6, color=color),
                text=labels,
                hoverinfo="text",
                hovertext=p_rel,
                textposition="top center",
                textfont=dict(size=16),
                showlegend=False,
            )
        )

        for subpath in kp.path:
            xyz = []
            for i in range(len(subpath)):
                xyz.append(kp.hs_coordinates[subpath[i]] @ cell)

            xyz = np.array(xyz).T
            self.fig.add_traces(
                data=go.Scatter3d(
                    mode="lines",
                    x=xyz[0],
                    y=xyz[1],
                    z=xyz[2],
                    line=dict(color=color),
                    hoverinfo="none",
                    showlegend=False,
                ),
            )


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir
